import numpy as np
import pytest

from fasr.data.audio import (
    Audio,
    AudioChannel,
    AudioChannelList,
    AudioChunk,
    AudioChunkList,
    AudioSpan,
    AudioSpanList,
    AudioToken,
    AudioTokenList,
)
from fasr.data.waveform import Waveform

# ==================== AudioToken ====================


class TestAudioToken:
    def test_create(self):
        token = AudioToken(text="你", start_ms=0, end_ms=100)
        assert token.text == "你"
        assert token.start_ms == 0
        assert token.end_ms == 100
        assert token.follow == " "

    def test_duration_ms(self, audio_token):
        assert audio_token.duration_ms == 100

    def test_is_punctuation(self):
        assert AudioToken(text="，", start_ms=0, end_ms=50).is_punctuation()
        assert AudioToken(text="。", start_ms=0, end_ms=50).is_punctuation()
        assert not AudioToken(text="你", start_ms=0, end_ms=50).is_punctuation()
        assert not AudioToken(text="a", start_ms=0, end_ms=50).is_punctuation()

    def test_default_follow(self):
        token = AudioToken(text="x", start_ms=0, end_ms=10)
        assert token.follow == " "

    def test_custom_follow(self):
        token = AudioToken(text="x", start_ms=0, end_ms=10, follow="")
        assert token.follow == ""


# ==================== AudioTokenList ====================


class TestAudioTokenList:
    def test_duration_ms(self, audio_token_list):
        assert audio_token_list.duration_ms == 400

    def test_duration(self, audio_token_list):
        assert audio_token_list.duration == 0.4

    def test_text(self, audio_token_list):
        assert audio_token_list.text == "你好世界"

    def test_text_with_follow(self):
        tokens = AudioTokenList(
            [
                AudioToken(text="hello", start_ms=0, end_ms=100, follow=" "),
                AudioToken(text="world", start_ms=100, end_ms=200, follow=" "),
            ]
        )
        assert tokens.text == "hello world"

    def test_empty_list(self):
        tokens = AudioTokenList()
        assert tokens.text == ""


# ==================== AudioSpan ====================


class TestAudioSpan:
    def test_create(self, audio_span):
        assert audio_span.start_ms == 0
        assert audio_span.end_ms == 1000

    def test_duration_ms(self, audio_span):
        assert audio_span.duration_ms == 1000

    def test_duration(self, audio_span):
        assert audio_span.duration == 1.0

    def test_text(self, audio_span):
        assert audio_span.text == "你好世界"

    def test_text_no_tokens(self):
        span = AudioSpan(start_ms=0, end_ms=1000, tokens=None)
        assert span.text == ""

    def test_text_setter(self):
        span = AudioSpan(start_ms=0, end_ms=1000)
        span.text = "test"
        assert span.text == "test"

    def test_sample_rate(self, audio_span, sample_rate):
        assert audio_span.sample_rate == sample_rate

    def test_sample_rate_no_waveform(self):
        span = AudioSpan(start_ms=0, end_ms=1000)
        assert span.sample_rate is None

    def test_lt(self):
        s1 = AudioSpan(start_ms=0, end_ms=500)
        s2 = AudioSpan(start_ms=0, end_ms=1000)
        assert s1 < s2
        assert not s2 < s1

    def test_gt(self):
        s1 = AudioSpan(start_ms=0, end_ms=1000)
        s2 = AudioSpan(start_ms=0, end_ms=500)
        assert s1 > s2
        assert not s2 > s1

    def test_getitem(self, audio_span):
        token = audio_span[0]
        assert isinstance(token, AudioToken)
        assert token.text == "你"

    def test_getitem_none_tokens(self):
        span = AudioSpan(start_ms=0, end_ms=1000, tokens=None)
        assert span[0] is None

    def test_len(self, audio_span):
        assert len(audio_span) == 4

    def test_len_no_tokens(self):
        span = AudioSpan(start_ms=0, end_ms=1000, tokens=None)
        assert len(span) == 0


# ==================== AudioSpanList ====================


class TestAudioSpanList:
    def test_duration_ms(self):
        spans = AudioSpanList(
            [
                AudioSpan(start_ms=0, end_ms=500),
                AudioSpan(start_ms=500, end_ms=1500),
            ]
        )
        assert spans.duration_ms == 1500

    def test_duration(self):
        spans = AudioSpanList(
            [
                AudioSpan(start_ms=0, end_ms=1000),
                AudioSpan(start_ms=1000, end_ms=2000),
            ]
        )
        assert spans.duration == 2.0

    def test_padded_duration_ms(self):
        spans = AudioSpanList(
            [
                AudioSpan(start_ms=0, end_ms=300),
                AudioSpan(start_ms=300, end_ms=1300),
            ]
        )
        assert spans.padded_duration_ms == 1000 * 2

    def test_padded_duration_ms_empty(self):
        spans = AudioSpanList()
        assert spans.padded_duration_ms == 0

    def test_max_duration_ms(self):
        spans = AudioSpanList(
            [
                AudioSpan(start_ms=0, end_ms=300),
                AudioSpan(start_ms=300, end_ms=1300),
            ]
        )
        assert spans.max_duration_ms == 1000

    def test_max_duration_ms_empty(self):
        spans = AudioSpanList()
        assert spans.max_duration_ms == 0

    def test_text(self):
        tokens1 = AudioTokenList(
            [AudioToken(text="你好", start_ms=0, end_ms=100, follow="")]
        )
        tokens2 = AudioTokenList(
            [AudioToken(text="世界", start_ms=100, end_ms=200, follow="")]
        )
        spans = AudioSpanList(
            [
                AudioSpan(start_ms=0, end_ms=500, tokens=tokens1),
                AudioSpan(start_ms=500, end_ms=1000, tokens=tokens2),
            ]
        )
        assert spans.text == "你好世界"


# ==================== AudioChannel ====================


class TestAudioChannel:
    def test_create(self, audio_channel):
        assert audio_channel.waveform is not None
        assert audio_channel.segments is not None

    def test_duration(self, audio_channel):
        assert abs(audio_channel.duration - 1.0) < 0.01

    def test_duration_ms(self, audio_channel):
        assert abs(audio_channel.duration_ms - 1000.0) < 10

    def test_start_end_ms(self, audio_channel):
        assert audio_channel.start_ms == 0
        assert abs(audio_channel.end_ms - 1000.0) < 10

    def test_text_from_segments(self, audio_channel):
        assert audio_channel.text == "你好世界"

    def test_text_from_raw_text(self, waveform):
        ch = AudioChannel(waveform=waveform, raw_text="override text")
        assert ch.text == "override text"

    def test_text_setter(self, audio_channel):
        audio_channel.text = "new text"
        assert audio_channel.text == "new text"
        assert audio_channel.raw_text == "new text"

    def test_text_empty(self, waveform):
        ch = AudioChannel(waveform=waveform)
        assert ch.text == ""

    def test_set_waveform(self, sample_rate):
        ch = AudioChannel()
        data = np.zeros(sample_rate, dtype=np.float32)
        ch.set_waveform(data, sample_rate)
        assert ch.waveform is not None
        assert ch.waveform.sample_rate == sample_rate

    def test_set_text(self):
        ch = AudioChannel()
        ch.set_text("hello")
        assert ch.text == "hello"

    def test_resample(self, audio_channel):
        audio_channel.resample(8000)
        assert audio_channel.waveform.sample_rate == 8000

    def test_getitem(self, audio_channel):
        span = audio_channel[0]
        assert isinstance(span, AudioSpan)

    def test_getitem_no_segments(self, waveform):
        ch = AudioChannel(waveform=waveform)
        assert ch[0] is None

    def test_len(self, audio_channel):
        assert len(audio_channel) == 1

    def test_len_no_segments(self, waveform):
        ch = AudioChannel(waveform=waveform)
        assert len(ch) == 0

    def test_add_token(self, waveform):
        ch = AudioChannel(waveform=waveform)
        ch.add_token("测", start_ms=None, end_ms=None)
        assert ch.tokens is not None
        assert len(ch.tokens) == 1
        assert ch.tokens[0].text == "测"

    def test_is_recognized_true(self, audio_channel):
        assert audio_channel.is_recognized() is True

    def test_is_recognized_false_no_segments(self, waveform):
        ch = AudioChannel(waveform=waveform)
        assert ch.is_recognized() is False

    def test_is_recognized_false_no_tokens(self, waveform):
        span = AudioSpan(start_ms=0, end_ms=500, waveform=waveform, tokens=None)
        ch = AudioChannel(waveform=waveform, segments=AudioSpanList([span]))
        assert ch.is_recognized() is False

    def test_text_from_tokens(self, waveform):
        ch = AudioChannel(waveform=waveform)
        ch.add_token("你", follow="")
        ch.add_token("好", follow=" ")
        assert ch.text == "你好 "

    def test_text_priority(self, waveform, audio_token_list):
        """raw_text > sents > segments > tokens"""
        segments = AudioSpanList(
            [AudioSpan(start_ms=0, end_ms=500, tokens=audio_token_list)]
        )
        ch = AudioChannel(waveform=waveform, segments=segments, raw_text="优先")
        assert ch.text == "优先"


# ==================== Audio ====================


class TestAudio:
    def test_create_empty(self):
        audio = Audio()
        assert audio.url is None
        assert audio.channels is None

    def test_create_with_url(self):
        audio = Audio(url="/path/to/audio.wav")
        assert audio.url == "/path/to/audio.wav"

    def test_path_to_str_validator(self, tmp_path):
        from pathlib import Path

        audio = Audio(url=Path("/some/path.wav"))
        assert isinstance(audio.url, str)

    def test_is_loaded_false(self):
        audio = Audio()
        assert audio.is_loaded is False

    def test_is_loaded_true(self, audio_with_channel):
        assert audio_with_channel.is_loaded is True

    def test_is_loaded_empty_channels(self):
        audio = Audio()
        audio.channels = AudioChannelList()
        assert audio.is_loaded is False

    def test_duration_s(self, audio_with_channel):
        assert abs(audio_with_channel.duration_s - 1.0) < 0.01

    def test_duration_ms(self, audio_with_channel):
        assert abs(audio_with_channel.duration_ms - 1000.0) < 10

    def test_append_channel(self):
        audio = Audio()
        audio.append_channel()
        assert audio.channels is not None
        assert len(audio.channels) == 1

    def test_append_channel_with_channel(self, waveform):
        audio = Audio()
        ch = AudioChannel(waveform=waveform)
        audio.append_channel(ch)
        assert len(audio.channels) == 1
        assert audio.channels[0].waveform is not None

    def test_getitem(self, audio_with_channel):
        ch = audio_with_channel[0]
        assert isinstance(ch, AudioChannel)

    def test_getitem_none(self):
        audio = Audio()
        assert audio[0] is None

    def test_len(self, audio_with_channel):
        assert len(audio_with_channel) == 1

    def test_len_no_channels(self):
        assert len(Audio()) == 0

    def test_audio_format(self):
        assert Audio(url="test.wav").audio_format == "wav"
        assert Audio(url="test.mp3").audio_format == "mp3"
        assert Audio(url="test.amr").audio_format == "amr"
        assert Audio(url="test.flac").audio_format == "flac"
        assert Audio(url="test.m4a").audio_format == "m4a"
        assert Audio().audio_format is None

    def test_set_hotwords(self, audio_with_channel):
        audio_with_channel.set_hotwords(["hello", "world"])
        assert audio_with_channel.hotwords == frozenset(["hello", "world"])

    def test_resample(self, audio_with_channel):
        audio_with_channel.resample(8000)
        assert audio_with_channel.channels[0].waveform.sample_rate == 8000

    def test_resample_no_channels(self):
        audio = Audio()
        audio.resample(8000)

    def test_clear_waveform(self, audio_with_channel):
        audio_with_channel.clear_waveform()
        assert audio_with_channel.channels[0].waveform is None

    def test_clear_text(self, waveform, sample_rate):
        ch = AudioChannel(waveform=waveform, raw_text="some text")
        audio = Audio(sample_rate=sample_rate)
        audio.channels = AudioChannelList([ch])
        audio.clear_text()
        assert audio.channels[0].raw_text is None

    def test_clear_bad(self):
        audio = Audio(is_bad=True, bad_reason="broken", bad_component="loader")
        audio.clear_bad()
        assert audio.is_bad is None
        assert audio.bad_reason is None
        assert audio.bad_component is None

    def test_chunk(self, audio_with_channel):
        chunks = audio_with_channel.chunk(chunk_size_ms=200)
        assert isinstance(chunks, AudioChunkList)
        assert len(chunks) > 0
        assert chunks[-1].is_last is True
        assert chunks[0].is_start is True

    def test_chunk_small_audio(self, sample_rate):
        data = np.zeros(sample_rate // 10, dtype=np.float32)
        wf = Waveform(data=data, sample_rate=sample_rate)
        ch = AudioChannel(waveform=wf)
        audio = Audio(sample_rate=sample_rate, duration=0.1)
        audio.channels = AudioChannelList([ch])
        chunks = audio.chunk(chunk_size_ms=200)
        assert len(chunks) == 1
        assert chunks[0].is_last is True
        assert chunks[0].is_start is True

    def test_chunk_no_channels_raises(self):
        audio = Audio()
        with pytest.raises(ValueError):
            audio.chunk(100)

    @pytest.mark.xfail(
        reason="AudioChannel 未定义 bad_component 字段，align_channel_bad 访问该属性会报错"
    )
    def test_align_channel_bad(self, waveform, sample_rate):
        ch = AudioChannel(
            waveform=waveform,
            is_bad=True,
            bad_reason="noise",
            bad_component="loader",
        )
        audio = Audio(sample_rate=sample_rate)
        audio.channels = AudioChannelList([ch])
        audio.align_channel_bad()
        assert audio.is_bad is True

    def test_align_channel_not_bad(self, waveform, sample_rate):
        ch = AudioChannel(waveform=waveform)
        audio = Audio(sample_rate=sample_rate)
        audio.channels = AudioChannelList([ch])
        audio.align_channel_bad()
        assert audio.is_bad is None


# ==================== AudioChunk ====================


class TestAudioChunk:
    def test_create(self, audio_chunk):
        assert audio_chunk.stream_id == "test-stream-id"
        assert audio_chunk.waveform is not None

    def test_text_empty(self, audio_chunk):
        assert audio_chunk.text == ""

    def test_text_with_tokens(self, waveform):
        tokens = AudioTokenList(
            [
                AudioToken(text="你", start_ms=0, end_ms=100, follow=""),
                AudioToken(text="好", start_ms=100, end_ms=200, follow=""),
            ]
        )
        chunk = AudioChunk(stream_id="s", waveform=waveform, tokens=tokens)
        assert chunk.text == "你好"

    def test_duration_ms(self, audio_chunk):
        assert abs(audio_chunk.duration_ms - 1000.0) < 10

    def test_duration_s(self, audio_chunk):
        assert abs(audio_chunk.duration_s - 1.0) < 0.01

    def test_is_final(self):
        wf = Waveform(data=np.zeros(100, dtype=np.float32), sample_rate=16000)
        chunk = AudioChunk(stream_id="s", waveform=wf, is_last=True)
        assert chunk.is_final is True

    def test_clear(self, audio_chunk):
        """AudioChunk.waveform 是必填字段，不能设为 None，验证 clear 会报错"""
        from pydantic import ValidationError

        with pytest.raises(ValidationError):
            audio_chunk.clear()


# ==================== AudioChunkList ====================


class TestAudioChunkList:
    def _make_chunks(self, sample_rate=16000):
        chunks = AudioChunkList()
        for i in range(3):
            data = np.zeros(sample_rate // 2, dtype=np.float32)
            wf = Waveform(data=data, sample_rate=sample_rate)
            chunks.append(
                AudioChunk(
                    stream_id="s",
                    waveform=wf,
                    start_ms=i * 500,
                    end_ms=(i + 1) * 500,
                )
            )
        return chunks

    def test_concat(self):
        chunks = self._make_chunks()
        audio = chunks.concat()
        assert isinstance(audio, Audio)
        assert audio.channels is not None
        assert len(audio.channels) == 1

    def test_concat_empty(self):
        chunks = AudioChunkList()
        audio = chunks.concat()
        assert isinstance(audio, Audio)
        assert audio.channels is None

    def test_concat_to_chunk(self):
        chunks = self._make_chunks()
        merged = chunks.concat_to_chunk()
        assert isinstance(merged, AudioChunk)
        assert merged.start_ms == 0
        assert merged.end_ms == 1500

    def test_concat_to_chunk_single(self):
        wf = Waveform(data=np.zeros(100, dtype=np.float32), sample_rate=16000)
        chunks = AudioChunkList([AudioChunk(stream_id="s", waveform=wf)])
        merged = chunks.concat_to_chunk()
        assert merged is chunks[0]

    def test_concat_to_audio(self):
        chunks = self._make_chunks()
        audio = chunks.concat_to_audio()
        assert isinstance(audio, Audio)
        assert audio.channels is not None


# ==================== AudioList ====================


class TestAudioList:
    def test_from_urls(self):
        from fasr.data.audio import AudioList

        urls = ["/path/a.wav", "/path/b.wav"]
        audios = AudioList.from_urls(urls, load=False)
        assert len(audios) == 2
        assert audios[0].url == "/path/a.wav"

    def test_from_urls_single(self):
        from fasr.data.audio import AudioList

        audios = AudioList.from_urls("/path/a.wav", load=False)
        assert len(audios) == 1

    def test_str(self):
        from fasr.data.audio import AudioList

        audios = AudioList([Audio(url="a.wav"), Audio(url="b.wav", is_bad=True)])
        s = str(audios)
        assert "2 audios" in s
        assert "1 bad" in s

    def test_bad_audios(self):
        from fasr.data.audio import AudioList

        audios = AudioList(
            [
                Audio(url="a.wav"),
                Audio(url="b.wav", is_bad=True),
                Audio(url="c.wav", is_bad=True),
            ]
        )
        assert len(audios.bad_audios) == 2
        assert len(audios.good_audios) == 1

    def test_has_bad_audio(self):
        from fasr.data.audio import AudioList

        audios = AudioList([Audio(url="a.wav", is_bad=True)])
        assert audios.has_bad_audio() is True

        audios2 = AudioList([Audio(url="a.wav")])
        assert audios2.has_bad_audio() is False

    def test_group_by_hotwords(self):
        from fasr.data.audio import AudioList

        a1 = Audio(url="a.wav", hotwords=frozenset(["cat"]))
        a2 = Audio(url="b.wav", hotwords=frozenset(["cat"]))
        a3 = Audio(url="c.wav", hotwords=frozenset(["dog"]))
        a4 = Audio(url="d.wav")
        audios = AudioList([a1, a2, a3, a4])
        groups = audios.group_by_hotwords()
        assert len(groups) == 3
        assert len(groups[frozenset(["cat"])]) == 2

    def test_copy(self, waveform, sample_rate):
        from fasr.data.audio import AudioList

        ch = AudioChannel(waveform=waveform)
        audio = Audio(sample_rate=sample_rate, duration=1.0)
        audio.channels = AudioChannelList([ch])
        audios = AudioList([audio])
        copied = audios.copy()
        assert len(copied) == 1
        assert copied[0] is not audios[0]

    def test_sub(self):
        from fasr.data.audio import AudioList

        a1 = Audio(url="a.wav")
        a2 = Audio(url="b.wav")
        list1 = AudioList([a1, a2])
        list2 = AudioList([a1])
        diff = list1 - list2
        assert len(diff) == 1



import numpy as np
import pytest

from fasr.data.audio import (
    Audio,
    AudioChannel,
    AudioChannelList,
    AudioSpan,
    AudioSpanList,
    AudioToken,
    AudioTokenList,
)
from fasr.data.eval import AudioEvalExample, AudioEvalResult
from fasr.data.waveform import Waveform


def _make_audio_with_text(text: str, sample_rate: int = 16000) -> Audio:
    data = np.zeros(sample_rate, dtype=np.float32)
    wf = Waveform(data=data, sample_rate=sample_rate)
    ch = AudioChannel(waveform=wf, raw_text=text)
    audio = Audio(sample_rate=sample_rate, duration=1.0)
    audio.channels = AudioChannelList([ch])
    return audio


class TestAudioEvalExample:
    def test_create(self):
        x = _make_audio_with_text("你好世界")
        y = _make_audio_with_text("你好世界")
        example = AudioEvalExample(x=x, y=y)
        assert example.is_badcase is False

    def test_channel_mismatch_raises(self):
        x = _make_audio_with_text("hello")
        y = Audio(sample_rate=16000)
        y.channels = AudioChannelList([AudioChannel(), AudioChannel()])
        with pytest.raises(ValueError):
            AudioEvalExample(x=x, y=y)

    def test_char_error_rate_perfect(self):
        x = _make_audio_with_text("你好世界")
        y = _make_audio_with_text("你好世界")
        example = AudioEvalExample(x=x, y=y)
        cer = example.char_error_rate()
        assert cer == 0.0

    def test_char_error_rate_different(self):
        x = _make_audio_with_text("你好啊世界")
        y = _make_audio_with_text("你好世界")
        example = AudioEvalExample(x=x, y=y)
        cer = example.char_error_rate()
        assert cer > 0
        assert example.is_badcase is True

    def test_clear_waveform(self):
        x = _make_audio_with_text("hello")
        y = _make_audio_with_text("hello")
        example = AudioEvalExample(x=x, y=y)
        example.clear_waveform()
        assert x.channels[0].waveform is None
        assert y.channels[0].waveform is None


class TestAudioEvalResult:
    def test_add_example(self):
        x = _make_audio_with_text("你好")
        y = _make_audio_with_text("你好")
        result = AudioEvalResult()
        result.add_example(x, y)
        assert len(result) == 1

    def test_add_example_list(self):
        from fasr.data.audio import AudioList

        x_list = AudioList(
            [_make_audio_with_text("你好"), _make_audio_with_text("世界")]
        )
        y_list = AudioList(
            [_make_audio_with_text("你好"), _make_audio_with_text("世界")]
        )
        result = AudioEvalResult()
        result.add_example(x_list, y_list)
        assert len(result) == 2

    def test_add_example_mismatched_raises(self):
        from fasr.data.audio import AudioList

        x_list = AudioList([_make_audio_with_text("你好")])
        y_list = AudioList(
            [_make_audio_with_text("你好"), _make_audio_with_text("世界")]
        )
        result = AudioEvalResult()
        with pytest.raises(ValueError):
            result.add_example(x_list, y_list)

    def test_char_error_rate(self):
        result = AudioEvalResult()
        result.add_example(
            _make_audio_with_text("你好世界"),
            _make_audio_with_text("你好世界"),
        )
        result.add_example(
            _make_audio_with_text("你好世界"),
            _make_audio_with_text("你好世界"),
        )
        cer = result.char_error_rate()
        assert cer == 0.0

    def test_x_y_properties(self):
        result = AudioEvalResult()
        result.add_example(
            _make_audio_with_text("你好"),
            _make_audio_with_text("世界"),
        )
        assert len(result.x) == 1
        assert len(result.y) == 1

    def test_get_badcase(self):
        result = AudioEvalResult()
        result.add_example(
            _make_audio_with_text("你好世界"),
            _make_audio_with_text("你好世界"),
        )
        result.add_example(
            _make_audio_with_text("完全不同"),
            _make_audio_with_text("你好世界"),
        )
        result.char_error_rate()
        badcases = result.get_badcase()
        assert len(badcases) == 1



from pathlib import Path
from random import choices
from typing import List


def random_hotwords(n: int) -> List[str]:
    """Generate random hotwords.

    Args:
        n (int): The number of hotwords.

    Returns:
        List[str]: The list of hotwords.
    """
    hotwords_file = Path(__file__).parent.parent / "asset" / "example" / "hotwords.txt"
    with open(hotwords_file, "r") as f:
        hotwords = f.readlines()
    hotwords = choices(hotwords, k=n)
    return [hotword.strip() for hotword in hotwords]



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Yan Jia)

from ..data.token_dict import TokenDict


class LidTokenizer:
    def __init__(self, dict_path, unk="<unk>"):
        self.dict = TokenDict(dict_path, unk=unk)

    def detokenize(self, inputs, join_symbol=" "):
        if len(inputs) > 0 and type(inputs[0]) == int:
            tokens = [self.dict[id] for id in inputs]
        else:
            tokens = inputs
        s = f"{join_symbol}".join(tokens)
        return s



import nest_asyncio

from .config import registry
from .data import Audio
from .pipeline import AudioPipeline
from .utils.base import get_cache_dir, set_cache_dir

nest_asyncio.apply()  # jupyter


def load(pipeline_name: str) -> AudioPipeline:
    from importlib.metadata import entry_points

    for plugin in entry_points(group="fasr.pipelines"):
        if plugin.name == pipeline_name:
            plugin.load()
            return registry.pipelines.get(pipeline_name)()
    raise ValueError(f"Pipeline {pipeline_name} not found")


__all__ = ["Audio", "AudioPipeline", "load", "set_cache_dir", "get_cache_dir"]



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Yan Jia)

from dataclasses import dataclass
import os
import time
import traceback

import torch

from .data.feat import FeatExtractor
from .models.fireredlid_aed import FireRedLidAed
from .models.param import count_model_parameters
from .tokenizer.lid_tokenizer import LidTokenizer


@dataclass
class FireRedLidConfig:
    use_gpu: bool = True
    use_half: bool = False


class FireRedLid:
    @classmethod
    def from_pretrained(cls, model_dir, config=FireRedLidConfig()):
        cmvn_path = os.path.join(model_dir, "cmvn.ark")
        feat_extractor = FeatExtractor(cmvn_path)

        model_path = os.path.join(model_dir, "model.pth.tar")
        dict_path = os.path.join(model_dir, "dict.txt")
        model = load_fireredlid_model(model_path)
        tokenizer = LidTokenizer(dict_path)

        count_model_parameters(model)
        model.eval()
        return cls(feat_extractor, model, tokenizer, config)

    def __init__(self, feat_extractor, model, tokenizer, config):
        self.feat_extractor = feat_extractor
        self.model = model
        self.tokenizer = tokenizer
        self.config = config
        self.config.beam_size = 3
        self.config.nbest = 1
        self.config.decode_max_len = 2
        self.config.softmax_smoothing = 1.25
        self.config.aed_length_penalty = 0.6
        self.config.eos_penalty = 1.0
        if self.config.use_gpu:
            if self.config.use_half:
                self.model.half()
            self.model.cuda()
        else:
            self.model.cpu()

    @torch.no_grad()
    def process(self, batch_uttid, batch_wav_path):
        batch_uttid_origin = batch_uttid
        try:
            feats, lengths, durs, batch_wav_path, batch_uttid = self.feat_extractor(
                batch_wav_path, batch_uttid
            )
            if feats is None:
                return [{"uttid": uttid, "lang": ""} for uttid in batch_uttid_origin]
        except Exception:
            traceback.print_exc()
            return [{"uttid": uttid, "lang": ""} for uttid in batch_uttid_origin]
        total_dur = sum(durs)
        if self.config.use_gpu:
            feats, lengths = feats.cuda(), lengths.cuda()
            if self.config.use_half:
                feats = feats.half()

        start_time = time.time()

        try:
            hyps = self.model.process(
                feats,
                lengths,
                self.config.beam_size,
                self.config.nbest,
                self.config.decode_max_len,
                self.config.softmax_smoothing,
                self.config.aed_length_penalty,
                self.config.eos_penalty,
            )
        except Exception:
            traceback.print_exc()
            hyps = []

        elapsed = time.time() - start_time
        rtf = elapsed / total_dur if total_dur > 0 else 0

        results = []
        for uttid, wav, hyp, dur in zip(batch_uttid, batch_wav_path, hyps, durs):
            hyp = hyp[0]
            hyp_ids = [int(id) for id in hyp["yseq"].cpu()]
            text = self.tokenizer.detokenize(hyp_ids)
            results.append(
                {
                    "uttid": uttid,
                    "lang": text,
                    "confidence": round(hyp["confidence"].cpu().item(), 3),
                    "dur_s": round(dur, 3),
                    "rtf": f"{rtf:.4f}",
                }
            )
            if type(wav) == str:
                results[-1]["wav"] = wav
        return results


def load_fireredlid_model(model_path):
    package = torch.load(
        model_path,
        map_location=lambda storage, loc: storage,
        weights_only=False,
    )
    model = FireRedLidAed.from_args(package["args"])
    model.load_state_dict(package["model_state_dict"], strict=False)
    return model



import ctypes

import numpy as np
import pytest

from fasr.data.audio import AudioChunk, AudioChunkList
from fasr.data.stream import AudioBytesStream


class TestAudioBytesStream:
    def test_init(self):
        stream = AudioBytesStream(sample_rate=16000, chunk_size_ms=100)
        assert stream._sample_rate == 16000
        expected_bytes = 16000 // 1000 * 100 * ctypes.sizeof(ctypes.c_int16)
        assert stream._bytes_per_frame == expected_bytes

    def test_push_exact_frame(self):
        sr = 16000
        chunk_ms = 100
        stream = AudioBytesStream(sample_rate=sr, chunk_size_ms=chunk_ms)

        samples_per_frame = sr // 1000 * chunk_ms
        data = np.zeros(samples_per_frame, dtype=np.int16).tobytes()
        chunks = stream.push(data)
        assert len(chunks) == 1
        assert isinstance(chunks[0], AudioChunk)

    def test_push_multiple_frames(self):
        sr = 16000
        chunk_ms = 100
        stream = AudioBytesStream(sample_rate=sr, chunk_size_ms=chunk_ms)

        samples_per_frame = sr // 1000 * chunk_ms
        data = np.zeros(samples_per_frame * 3, dtype=np.int16).tobytes()
        chunks = stream.push(data)
        assert len(chunks) == 3

    def test_push_partial_then_complete(self):
        sr = 16000
        chunk_ms = 100
        stream = AudioBytesStream(sample_rate=sr, chunk_size_ms=chunk_ms)

        samples_per_frame = sr // 1000 * chunk_ms
        half = np.zeros(samples_per_frame // 2, dtype=np.int16).tobytes()

        chunks = stream.push(half)
        assert len(chunks) == 0

        chunks = stream.push(half)
        assert len(chunks) == 1

    def test_push_incremental(self):
        sr = 16000
        chunk_ms = 100
        stream = AudioBytesStream(sample_rate=sr, chunk_size_ms=chunk_ms)

        samples_per_frame = sr // 1000 * chunk_ms
        total_chunks = 0
        for _ in range(10):
            small = np.zeros(samples_per_frame // 5, dtype=np.int16).tobytes()
            chunks = stream.push(small)
            total_chunks += len(chunks)
        assert total_chunks == 2

    def test_flush_remaining(self):
        """flush 时 buffer 是 bytearray 而非 ndarray，会导致 Waveform 验证失败（已知 bug）"""
        sr = 16000
        chunk_ms = 100
        stream = AudioBytesStream(sample_rate=sr, chunk_size_ms=chunk_ms)

        samples = sr // 1000 * chunk_ms + 100
        data = np.zeros(samples, dtype=np.int16).tobytes()
        chunks = stream.push(data)
        assert len(chunks) == 1

        from pydantic import ValidationError

        with pytest.raises(ValidationError):
            stream.flush()

    def test_flush_empty(self):
        stream = AudioBytesStream(sample_rate=16000, chunk_size_ms=100)
        chunks = stream.flush()
        assert len(chunks) == 0

    def test_stream_id_consistent(self):
        sr = 16000
        chunk_ms = 100
        stream = AudioBytesStream(sample_rate=sr, chunk_size_ms=chunk_ms)

        samples = sr // 1000 * chunk_ms * 3
        data = np.zeros(samples, dtype=np.int16).tobytes()
        chunks = stream.push(data)

        ids = {c.stream_id for c in chunks}
        assert len(ids) == 1

    def test_waveform_dtype(self):
        sr = 16000
        chunk_ms = 100
        stream = AudioBytesStream(sample_rate=sr, chunk_size_ms=chunk_ms)

        samples = sr // 1000 * chunk_ms
        data = np.ones(samples, dtype=np.int16).tobytes()
        chunks = stream.push(data)
        assert chunks[0].waveform.data.dtype == np.float32

    def test_multichannel(self):
        sr = 16000
        chunk_ms = 100
        num_channels = 2
        stream = AudioBytesStream(
            sample_rate=sr, num_channels=num_channels, chunk_size_ms=chunk_ms
        )

        samples = sr // 1000 * chunk_ms * num_channels
        data = np.zeros(samples, dtype=np.int16).tobytes()
        chunks = stream.push(data)
        assert len(chunks) == 1



from pathlib import Path
from typing import Literal

from jsonargparse import auto_cli


class Benchmark:
    """基准测试类"""

    def __init__(self):
        pass

    def loader(self, urls: str):
        """测试fasr的loader性能.

        Args:
            urls (str): url文件路径, 每行一个url.
        """
        import time

        from loguru import logger

        from fasr.components import AudioLoader

        loader = AudioLoader()

        if not Path(urls).exists():
            raise FileNotFoundError(f"{urls} not found")
        if Path(urls).is_dir():
            urls = [
                str(p)
                for p in Path(urls).iterdir()
                if p.is_file() and p.suffix == ".wav"
            ]
        else:
            with open(urls, "r") as f:
                urls = f.readlines()
            urls = [url.strip() for url in urls]

        # warm up
        _ = urls[:5] | loader

        # benchmark
        start = time.perf_counter()
        audios = urls | loader
        end = time.perf_counter()
        took = round(end - start, 2)
        duration = audios.duration_s
        logger.info(f"took {took} seconds, speedup: {round(duration / took, 2)}")

    def vad(
        self,
        urls: str,
        model: str = "fsmn",
        batch_size: int = 2,
        num_threads: int = 2,
        num_samples: int | None = None,
        audio_format: Literal["wav", "mp3"] = "wav",
    ):
        """对比测试fasr与funasr pipeline(load -> vad)的性能.

        Args:
            urls (str): url文件路径, 每行一个url.
            batch_size (int, optional): 批处理大小. Defaults to 2.
            num_threads (int, optional): 并行处理的工作线程数. Defaults to 2.
            num_samples (int, optional): 采样数量. Defaults to 100.
            model_dir (str, optional): 语音检测模型目录. Defaults to "checkpoints/iic/speech_fsmn_vad_zh-cn-16k-common-pytorch".
        """
        import time

        from loguru import logger
        from tqdm import trange

        from fasr.components import VoiceDetector
        from fasr.data import AudioList

        vad = VoiceDetector().setup(model=model, num_threads=num_threads)

        if not Path(urls).exists():
            raise FileNotFoundError(f"{urls} not found")
        if Path(urls).is_dir():
            urls = [
                str(p)
                for p in Path(urls).iterdir()
                if p.is_file() and p.suffix == f".{audio_format}"
            ]
        else:
            with open(urls, "r") as f:
                urls = f.readlines()
            urls = [url.strip() for url in urls]
        if num_samples:
            urls = urls[:num_samples]

        audios = AudioList()
        duration = 0
        for idx in range(0, len(urls), batch_size):
            batch_urls = urls[idx : idx + batch_size]
            batch_audios = AudioList.from_urls(
                urls=batch_urls, load=True, num_workers=4
            )
            for audio in batch_audios:
                duration += audio.duration * len(audio.channels)
            audios.extend(batch_audios)

        def run_vad(audios: AudioList) -> float:
            start = time.perf_counter()
            for i in trange(0, len(audios), batch_size):
                _audios = audios[i : i + batch_size]
                _audios = vad.predict(_audios)
            end = time.perf_counter()
            took = round(end - start, 2)
            return took

        # warm up
        logger.info("warm up")
        _ = run_vad(audios[0:1])

        # benchmark
        fasr_took = run_vad(audios)
        logger.info(f"All channels duration: {round(duration, 2)} seconds")
        logger.info(
            f"{model}: took {fasr_took} seconds, speedup: {round(duration / fasr_took, 2)}"
        )

    def stream_vad(
        self,
        model: str,
        max_silence: int = 500,
        threshold: float = 0.8,
        chunk_size_ms: int = 200,
        db_threshold: int = -100,
        print_infer_time: bool = False,
        interactive: bool = False,
    ):
        """流式语音活动检测"""
        import time

        from loguru import logger

        from fasr.config import registry
        from fasr.data import Audio, AudioBytesStream, AudioChunk, AudioSpan
        from fasr.model import StreamVADModel

        vad: StreamVADModel = registry.stream_vad_models.get(model)(
            chunk_size_ms=chunk_size_ms,
            max_end_silence_time=max_silence,
            speech_noise_thres=threshold,
            db_threshold=db_threshold,
        ).from_checkpoint()

        if interactive:
            import pyaudio

            FORMAT = pyaudio.paInt16
            RATE = 16000
            chunk_size_ms = chunk_size_ms
            CHUNK = int(RATE / 1000 * chunk_size_ms)

            p = pyaudio.PyAudio()
            stream = p.open(
                format=FORMAT,
                channels=1,
                rate=RATE,
                input=True,
                frames_per_buffer=CHUNK,
            )
            bytes_buffer = AudioBytesStream(
                sample_rate=RATE, chunk_size_ms=chunk_size_ms
            )
            logger.info("开始录音...")
            while True:
                bytes_data = stream.read(CHUNK)
                chunks = bytes_buffer.push(bytes_data)
                for chunk in chunks:
                    chunk: AudioChunk
                    if print_infer_time:
                        start = time.perf_counter()
                    for span in vad.detect_chunk(chunk=chunk):
                        logger.info(f"speech {span.vad_state}")
                    if print_infer_time:
                        end = time.perf_counter()
                        spend = round(end - start, 4)
                        duration = chunk.duration_s
                        speedup = round(duration / spend, 4)
                        rtf = round(spend / duration, 4)
                        logger.info(
                            f"chunk duration: {duration}, spend: {spend}, speedup: {speedup}, rtf: {rtf}"
                        )
        else:
            audio = Audio().load_example(example="asr")
            chunks = audio.chunk(chunk_size_ms=chunk_size_ms)
            spent_times = []
            for chunk in chunks:
                chunk: AudioChunk
                start = time.perf_counter()
                for span in vad.detect_chunk(chunk):
                    end = time.perf_counter()
                    spend = end - start
                    span: AudioSpan
                    if span.vad_state == "start":
                        logger.info(f"speech start: {span.start_ms}")
                    elif span.vad_state == "end":
                        logger.info(f"speech end: {span.end_ms}")
                    spent_times.append(spend)
            mean_spend = round(sum(spent_times) / len(spent_times), 5)
            speedup = round(chunk_size_ms / mean_spend / 1000, 4)
            rtf = round(mean_spend / (chunk_size_ms / 1000), 4)
            logger.info(
                f"chunk size: {chunk_size_ms} ms, mean spend: {mean_spend} s, speedup: {speedup}, rtf: {rtf}"
            )

    def asr(
        self,
        urls: str,
        model: str,
        batch_size: int = 2,
        num_threads: int = 1,
        num_samples: int | None = None,
    ):
        """测试asr推理性能.

        Args:
            urls (str): url文件路径.
            batch_size (int, optional): 批处理大小. Defaults to 2.
            num_threads (int, optional): 并行处理的工作线程数. Defaults to 2.
            num_samples (int, optional): 采样数量. Defaults to 100.
            model_dir (str, optional): 语音检测模型目录. Defaults to "checkpoints/iic/speech_fsmn_vad_zh-cn-16k-common-pytorch".
        """
        import time

        from loguru import logger
        from tqdm import trange

        from fasr.components import AudioLoader, SpeechRecognizer
        from fasr.data import Audio, AudioList

        asr = SpeechRecognizer().setup(model=model, num_threads=num_threads)
        loader = AudioLoader()

        if not Path(urls).exists():
            raise FileNotFoundError(f"{urls} not found")
        if Path(urls).is_dir():
            urls = [
                str(p)
                for p in Path(urls).iterdir()
                if p.is_file() and p.suffix == ".wav"
            ]
        else:
            with open(urls, "r") as f:
                urls = f.readlines()
            urls = [url.strip() for url in urls]
        if num_samples:
            urls = urls[:num_samples]

        audios = AudioList()
        duration = 0
        for idx in range(0, len(urls), batch_size):
            batch_urls = urls[idx : idx + batch_size]
            batch_audios = AudioList.from_urls(urls=batch_urls)
            batch_audios = loader.predict(batch_audios)
            for audio in batch_audios:
                duration += audio.duration * len(audio.channels)
            audios.extend(batch_audios)

        def run_asr(audios: AudioList) -> float:
            start = time.perf_counter()
            for i in trange(0, len(audios), batch_size):
                _audios = audios[i : i + batch_size]
                _audios = asr.predict(_audios)
            end = time.perf_counter()
            took = round(end - start, 2)
            return took

        # warm up
        logger.info("warm up")
        _ = run_asr(audios[0:1])

        # benchmark
        fasr_took = run_asr(audios)
        logger.info(f"All channels duration: {round(duration, 2)} seconds")
        logger.info(
            f"{model}: took {fasr_took} seconds, speedup: {round(duration / fasr_took, 2)}"
        )

    def stream_asr(
        self,
        model: str,
        chunk_size_ms: int = 600,
        interactive: bool = False,
        compile: bool = False,
        device: str = None,
    ):
        """流式语音识别"""
        import time

        from loguru import logger

        from fasr.config import registry
        from fasr.data import Audio, AudioBytesStream, AudioChunk
        from fasr.model import StreamASRModel

        asr: StreamASRModel = registry.stream_asr_models.get(model)(
            chunk_size_ms=chunk_size_ms
        ).from_checkpoint(compile=compile, device=device)
        audio = Audio().load_example(example="asr")
        chunks = audio.chunk(chunk_size_ms=chunk_size_ms)
        if compile:
            logger.info("warm up model")
            for chunk in chunks:
                chunk: AudioChunk
                for token in asr.transcribe_chunk(chunk=chunk):
                    pass
        if interactive:
            import pyaudio

            FORMAT = pyaudio.paInt16
            RATE = 16000
            chunk_size_ms = chunk_size_ms
            CHUNK = int(RATE / 1000 * chunk_size_ms)
            p = pyaudio.PyAudio()
            stream = p.open(
                format=FORMAT,
                channels=1,
                rate=RATE,
                input=True,
                frames_per_buffer=CHUNK,
            )
            bytes_buffer = AudioBytesStream(
                sample_rate=RATE, chunk_size_ms=chunk_size_ms
            )
            logger.info("开始录音...")
            while True:
                bytes_data = stream.read(CHUNK)
                chunks = bytes_buffer.push(bytes_data)
                start = time.perf_counter()
                text = ""
                for chunk in chunks:
                    chunk: AudioChunk
                    for token in asr.transcribe_chunk(chunk=chunk):
                        text += token.text
                if text:
                    logger.info(text)
                    end = time.perf_counter()
                    spend = round(end - start, 4)
                    duration = chunk_size_ms / 1000
                    speedup = round(duration / spend, 4)
                    rtf = round(spend / duration, 4)
                    logger.info(
                        f"chunk size: {chunk_size_ms}ms, spend: {spend}, speedup: {speedup}, rtf: {rtf}"
                    )

        else:
            chunk_spends = []
            for chunk in chunks:
                start = time.perf_counter()
                chunk: AudioChunk
                text = ""
                for token in asr.transcribe_chunk(chunk=chunk):
                    text += token.text
                if text:
                    logger.info(text)
                    end = time.perf_counter()
                    chunk_spends.append(end - start)
            mean_spend = round(sum(chunk_spends) / len(chunk_spends), 5)
            speedup = round(chunk_size_ms / mean_spend / 1000, 4)
            rtf = round(mean_spend / (chunk_size_ms / 1000), 4)
            logger.info(
                f"chunk size: {chunk_size_ms} ms, mean spend: {mean_spend} s, speedup: {speedup}, rtf: {rtf}"
            )

    def pipeline(
        self,
        name: str,
        input: str,
        batch_size: int = 2,
        num_threads: int = 2,
        batch_size_s: int = 100,
        num_samples: int | None = None,
    ):
        """对比测试fasr与funasr pipeline(load -> vad -> asr -> punc)的性能.

        Args:
            input (str): 测试文件，格式可以为一行为一个url的txt文件可以为AudioList保存格式.
            batch_size (int, optional): 批处理大小. Defaults to 2.
            batch_size_s (int, optional): asr模型的批处理大小. Defaults to 100.
            num_threads (int, optional): 并行处理的工作线程数. Defaults to 2.
            num_samples (int, optional): 采样数量. Defaults to 100.
            asr_model (str, optional): asr模型. Defaults to "paraformer".
            vad_model (str, optional): vad模型. Defaults to "fsmn".
            punc_model (str, optional): 标点模型. Defaults to "ct_transformer".
        """
        import time

        from loguru import logger

        import fasr
        from fasr.data import AudioList

        def check_input_and_load(input: str) -> AudioList:
            """检查输入文件是否存在.

            Args:
                input (str): 输入文件路径.
            """
            if not Path(input).exists():
                raise FileNotFoundError(f"{input} not found")
            if Path(input).is_dir():
                files = [
                    str(p)
                    for p in Path(input).iterdir()
                    if p.is_file() and p.suffix == ".wav"
                ]
                audios = AudioList.from_urls(files)
            else:
                if Path(input).suffix == ".al":
                    audios = AudioList.load_binary(input)
                elif Path(input).suffix == ".txt":
                    with open(input, "r") as f:
                        urls = f.readlines()
                        urls = [url.strip() for url in urls]
                        audios = AudioList.from_urls(urls)
            return audios

        audios = check_input_and_load(input)
        audios = audios.load(num_workers=4)
        duration = audios.duration_s
        if num_samples:
            audios = audios[:num_samples]
        asr = fasr.load(name)

        def run_pipeline(urls):
            start = time.perf_counter()
            _ = asr.run(urls, verbose=True)
            end = time.perf_counter()
            took = round(end - start, 2)
            return took

        # warm up
        logger.info("warm up")
        _ = run_pipeline(audios[0:2])

        # benchmark
        logger.info("benchmark")
        pipeline_took = run_pipeline(audios)
        # 所有通道的总时长
        logger.info(f"duration: {round(duration, 2)} seconds")
        logger.info(
            f"pipeline: took {pipeline_took} seconds, speedup: {round(duration / pipeline_took, 2)}"
        )


class Serve:
    """服务类"""

    def __init__(self):
        pass

    def online(
        self,
        host: str = "0.0.0.0",
        port: int = 27000,
        device: str | None = None,
        vad_model: Literal[
            "stream_fsmn.torch", "stream_fsmn.onnx"
        ] = "stream_fsmn.torch",
        vad_chunk_size_ms: int = 100,
        vad_end_silence_ms: int = 500,
        vad_threshold: float = 0.6,
        vad_db_threshold: int = -100,
        asr_model: Literal[
            "stream_sensevoice", "stream_paraformer.onnx", "stream_paraformer.torch"
        ] = "stream_paraformer.torch",
        asr_checkpoint_dir: str | None = None,
        punc_model: Literal["ct_transformer"] | None = None,
        punc_checkpoint_dir: str | None = None,
    ):
        """流式语音识别"""
        from fasr.service.online import RealtimeASRService

        service = RealtimeASRService(
            host=host,
            port=port,
            device=device,
            asr_model_name=asr_model,
            asr_checkpoint_dir=asr_checkpoint_dir,
            vad_model_name=vad_model,
            vad_chunk_size_ms=vad_chunk_size_ms,
            vad_end_silence_ms=vad_end_silence_ms,
            vad_threshold=vad_threshold,
            vad_db_threshold=vad_db_threshold,
            punc_model_name=punc_model,
            punc_checkpoint_dir=punc_checkpoint_dir,
        )
        service.setup()

    def offline(self):
        """离线语音识别"""
        pass


commands = {
    "benchmark": Benchmark,
    "serve": Serve,
}


def run():
    """命令行"""
    auto_cli(components=commands)


if __name__ == "__main__":
    run()



from pathlib import Path

import numpy as np
from pydantic import Field, model_validator
import torch
from torch import Tensor
from torchaudio.compliance.kaldi import fbank as torchaudio_fbank

from fasr.config import Config
from fasr.data import Waveform
from fasr.utils import read_yaml

from .base import BaseWaveformPreprocessor


class FbankExtractor(BaseWaveformPreprocessor):
    """Conventional frontend structure for ASR."""

    fs: int = 16000
    window: str = "hamming"
    n_mels: int = 80
    frame_length: int = 25
    frame_shift: int = 10
    lfr_m: int = 5
    lfr_n: int = 1
    dither: float = 0.0
    round_to_power_of_two: bool = True
    snip_edges: bool = True
    cmvn_tensor: Tensor | np.ndarray | None = Field(
        None, description="CMVN tensor", title="CMVN tensor"
    )
    compile: bool = False

    @model_validator(mode="after")
    def compile_fbank(self):
        if self.compile:
            import torch._dynamo.config

            torch._dynamo.config.suppress_errors = True
            torch._dynamo.config.numpy_default_float = "float32"
            global torchaudio_fbank
            torchaudio_fbank = torch.compile(torchaudio_fbank)
            data = np.random.randn(1000)
            waveform = Waveform(data=data, sample_rate=self.fs)
            _ = self.process_waveform(waveform=waveform)
        return self

    def process_waveform(self, waveform: Waveform, device: str = "cpu") -> Waveform:
        if waveform.sample_rate != self.fs:
            waveform = waveform.resample(self.fs)
        data = waveform.data
        with torch.device(device):
            feats = self.apply_fbank(data)
        if self.lfr_m != 1 or self.lfr_n != 1:
            feats = self.apply_lfr(feats)
        if self.cmvn_tensor is not None:
            feats = self.apply_cmvn(feats)
        waveform.feats = feats
        return waveform

    def apply_fbank(self, waveform: np.ndarray) -> np.ndarray:
        waveform = waveform * (1 << 15)
        waveform = torch.from_numpy(waveform)
        feat = torchaudio_fbank(
            waveform.unsqueeze(0),
            channel=0,
            num_mel_bins=self.n_mels,
            window_type=self.window,
            frame_length=self.frame_length,
            frame_shift=self.frame_shift,
            round_to_power_of_two=self.round_to_power_of_two,
            dither=self.dither,
            sample_frequency=self.fs,
            energy_floor=0,
            snip_edges=self.snip_edges,
        )
        return feat.cpu().numpy().astype(np.float32)

    def apply_lfr(self, inputs: np.ndarray) -> np.ndarray:
        """低帧率技术"""
        LFR_inputs = []
        lfr_m, lfr_n = self.lfr_m, self.lfr_n
        T = inputs.shape[0]
        T_lfr = int(np.ceil(T / lfr_n))
        left_padding = np.tile(inputs[0], ((lfr_m - 1) // 2, 1))
        inputs = np.vstack((left_padding, inputs))
        T = T + (lfr_m - 1) // 2
        for i in range(T_lfr):
            if lfr_m <= T - i * lfr_n:
                LFR_inputs.append(
                    (inputs[i * lfr_n : i * lfr_n + lfr_m]).reshape(1, -1)
                )
            else:
                # process last LFR frame
                num_padding = lfr_m - (T - i * lfr_n)
                frame = inputs[i * lfr_n :].reshape(-1)
                for _ in range(num_padding):
                    frame = np.hstack((frame, inputs[-1]))

                LFR_inputs.append(frame)
        LFR_outputs = np.vstack(LFR_inputs).astype(np.float32)
        return LFR_outputs

    def apply_cmvn(self, inputs: np.ndarray) -> np.ndarray:
        """
        Apply CMVN with mvn data
        """
        frame, dim = inputs.shape
        means = np.tile(self.cmvn_tensor[0:1, :dim], (frame, 1))
        vars = np.tile(self.cmvn_tensor[1:2, :dim], (frame, 1))
        inputs = (inputs + means) * vars
        return inputs

    def warmup(self) -> None:
        waveform = np.random.randn(10000).astype(np.float32)
        self.process_waveform(waveform)

    def from_checkpoint(self, checkpoint_dir: str, **kwargs) -> "FbankExtractor":
        config: dict = read_yaml(Path(checkpoint_dir, "config.yaml"))["frontend_conf"]
        for key, value in config.items():
            if hasattr(self, key):
                setattr(self, key, value)
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
        cmvn_file = str(Path(checkpoint_dir, "am.mvn"))
        cmvn_tensor = load_cmvn_tensor(cmvn_file)
        self.cmvn_tensor = cmvn_tensor
        return self

    def get_config(self) -> Config:
        data = {
            "audio_preprocessor": {
                "@audio_preprocessors": "fbank_extractor",
                "fs": self.fs,
                "window": self.window,
                "n_mels": self.n_mels,
                "frame_length": self.frame_length,
                "frame_shift": self.frame_shift,
                "lfr_m": self.lfr_m,
                "lfr_n": self.lfr_n,
                "dither": self.dither,
                "round_to_power_of_two": self.round_to_power_of_two,
                "compile": self.compile,
            }
        }
        return Config(data=data)

    def save(self, save_dir: str) -> None:
        save_dir = Path(save_dir)
        if not save_dir.exists():
            save_dir.mkdir(parents=True)
        config_path = save_dir / "config.cfg"
        cmvn_path = save_dir / "cmvn.npy"
        np.save(cmvn_path, self.cmvn_tensor)
        self.get_config().to_disk(config_path)

    def load(self, save_dir: str) -> "FbankExtractor":
        save_dir = Path(save_dir)
        cmvn_path = save_dir / "cmvn.npy"
        self.cmvn_tensor = np.load(cmvn_path)
        return self


def low_frame_rate(inputs: Tensor, lfr_m: Tensor, lfr_n: Tensor) -> Tensor:
    LFR_inputs = []
    T = inputs.shape[0]
    T_lfr = int(np.ceil(T / lfr_n))
    left_padding = inputs[0].repeat((lfr_m - 1) // 2, 1)
    inputs = torch.vstack((left_padding, inputs))
    T = T + (lfr_m - 1) // 2
    for i in range(T_lfr):
        if lfr_m <= T - i * lfr_n:
            LFR_inputs.append((inputs[i * lfr_n : i * lfr_n + lfr_m]).view(1, -1))
        else:  # process last LFR frame
            num_padding = lfr_m - (T - i * lfr_n)
            frame = (inputs[i * lfr_n :]).view(-1)
            for _ in range(num_padding):
                frame = torch.hstack((frame, inputs[-1]))
            LFR_inputs.append(frame)
    LFR_outputs = torch.vstack(LFR_inputs)
    return LFR_outputs.type(torch.float32)


def load_cmvn_tensor(
    cmvn_file: str,
) -> Tensor:
    with open(cmvn_file, "r", encoding="utf-8") as f:
        lines = f.readlines()

    means_list = []
    vars_list = []
    for i in range(len(lines)):
        line_item = lines[i].split()
        if line_item[0] == "<AddShift>":
            line_item = lines[i + 1].split()
            if line_item[0] == "<LearnRateCoef>":
                add_shift_line = line_item[3 : (len(line_item) - 1)]
                means_list = list(add_shift_line)
                continue
        elif line_item[0] == "<Rescale>":
            line_item = lines[i + 1].split()
            if line_item[0] == "<LearnRateCoef>":
                rescale_line = line_item[3 : (len(line_item) - 1)]
                vars_list = list(rescale_line)
                continue

    means = np.array(means_list).astype(np.float64)
    vars = np.array(vars_list).astype(np.float64)
    cmvn = np.array([means, vars]).astype(np.float32)
    return cmvn



from typing import List

from joblib import Parallel, delayed
from loguru import logger
from pydantic import Field
from typing_extensions import Self

from fasr.components.base import BaseComponent
from fasr.config import registry
from fasr.data import AudioChannel, AudioList
from fasr.model import VADModel


@registry.components.register("detector")
class VoiceDetector(BaseComponent):
    name: str = "detector"
    description: str = "Detect voice segments from audio channels"
    input_tags: List[str] = ["Audio.channels"]
    output_tags: List[str] = ["AudioChannel.segments"]

    model: VADModel | None = Field(
        None, description="The vad model to use for voice detection"
    )
    threshold: float | None = Field(
        None,
        description="The threshold for audio duration, only audios with duration > threshold will be detected",
    )
    num_threads: int | None = Field(
        None, description="The number of threads to use for parallel processing"
    )

    def predict(self, audios: AudioList) -> AudioList:
        _audios = self.check_audios_threshold(audios)
        all_channels = [channel for audio in _audios for channel in audio.channels]
        num_threads = self.num_threads or 2
        batch_size = max(1, len(all_channels) // num_threads)
        _ = Parallel(n_jobs=num_threads, prefer="threads", batch_size=batch_size)(
            delayed(self.predict_channel)(channel) for channel in all_channels
        )
        return audios

    def predict_channel(self, channel: AudioChannel) -> AudioChannel:
        segments = self.model.detect(waveform=channel.waveform)
        channel.segments = segments
        return channel

    def check_audios_threshold(self, audios: AudioList) -> AudioList:
        # 大于threshold的音频才进行检测
        if self.threshold is not None:
            new_audios = AudioList()
            for audio in audios:
                if audio.duration > self.threshold:
                    new_audios.append(audio)
                else:
                    logger.warning(
                        f"Audio {audio.id} duration {audio.duration} < threshold {self.threshold}, skip detection"
                    )
            return new_audios
        return audios

    def setup(
        self,
        model: str,
        num_threads: int | None = None,
        threshold: float = None,
        **kwargs,
    ) -> Self:
        self.load_model_plugin(group="fasr.vad_models", model=model)
        self.model: VADModel = registry.vad_models.get(model)()
        self.model = self.model.from_checkpoint(**kwargs)
        self.num_threads = num_threads
        self.threshold = threshold
        return self



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu)

import glob
import logging
import os

logger = logging.getLogger(__name__)

from textgrid import IntervalTier, TextGrid


def get_wav_info(args):
    """
    Returns:
        wavs: list of (uttid, wav_path)
    """
    base = lambda p: os.path.basename(p).replace(".wav", "")
    if args.wav_path:
        wavs = [(base(args.wav_path), args.wav_path)]
    elif args.wav_paths and len(args.wav_paths) >= 1:
        wavs = [(base(p), p) for p in sorted(args.wav_paths)]
    elif args.wav_scp:
        wavs = [line.strip().split() for line in open(args.wav_scp)]
        if args.sort_wav_by_dur:
            logger.info("Sort wav by duration...")
            utt2dur = os.path.join(os.path.dirname(args.wav_scp), "utt2dur")
            if os.path.exists(utt2dur):
                utt2dur = [l.strip().split() for l in open(utt2dur)]
                utt2dur = {l[0]: float(l[1]) for l in utt2dur if len(l) == 2}
                wavs = sorted(wavs, key=lambda x: -utt2dur[x[0]])
                logger.info("Sort Done")
            else:
                logger.info(f"Not find {utt2dur}, un-sort")
    elif args.wav_dir:
        wavs = glob.glob(f"{args.wav_dir}/**/*.wav", recursive=True)
        wavs = [(base(p), p) for p in sorted(wavs)]
    else:
        raise ValueError("Please provide valid wav info")
    logger.info(f"#wavs={len(wavs)}")
    return wavs


def write_textgrid(wav_path, wav_dur, event):
    textgrid_file = wav_path.replace(".wav", ".TextGrid")
    logger.info(f"Write {textgrid_file}")
    textgrid = TextGrid(maxTime=wav_dur)
    tier = IntervalTier(name="token", maxTime=wav_dur)
    for token, start_s, end_s in event:
        if start_s == end_s:
            logger.info(f"Write TG, skip start=end {start_s}")
            continue
        start_s = max(start_s, 0)
        end_s = min(end_s, wav_dur)
        tier.add(minTime=start_s, maxTime=end_s, mark=token)
    textgrid.append(tier)
    textgrid.write(textgrid_file)



#!/usr/bin/env python3

# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu)

import argparse
from collections import OrderedDict
import re

parser = argparse.ArgumentParser()
parser.add_argument("--ref", type=str, required=True)
parser.add_argument("--hyp", type=str, required=True)
parser.add_argument("--print_sentence_wer", type=int, default=0)
parser.add_argument("--do_tn", type=int, default=0, help="simple tn by cn2an")
parser.add_argument("--rm_special", type=int, default=0, help="remove <\|.*?\|>")


def main(args):
    uttid2refs = read_uttid2tokens(args.ref, args.do_tn, args.rm_special)
    uttid2hyps = read_uttid2tokens(args.hyp, args.do_tn, args.rm_special)
    uttid2wer_info, wer_stat, en_dig_stat = compute_uttid2wer_info(
        uttid2refs, uttid2hyps, args.print_sentence_wer
    )
    wer_stat.print()
    en_dig_stat.print()


def read_uttid2tokens(filename, do_tn=False, rm_special=False):
    print(f">>> Read uttid to tokens: {filename}", flush=True)
    uttid2tokens = OrderedDict()
    uttid2text = read_uttid2text(filename, do_tn, rm_special)
    for uttid, text in uttid2text.items():
        tokens = text2tokens(text)
        uttid2tokens[uttid] = tokens
    return uttid2tokens


def read_uttid2text(filename, do_tn=False, rm_special=False):
    uttid2text = OrderedDict()
    with open(filename, "r", encoding="utf8") as fin:
        for i, line in enumerate(fin):
            cols = line.split()
            if len(cols) == 0:
                print("[WARN] empty line, continue", i, flush=True)
                continue
            assert cols[0] not in uttid2text, f"repeated uttid: {line}"
            if len(cols) == 1:
                uttid2text[cols[0]] = ""
                continue
            txt = " ".join(cols[1:])
            if rm_special:
                txt = " ".join(
                    [t for t in re.split("<\|.*?\|>", txt) if t.strip() != ""]
                )
            if do_tn:
                import cn2an

                txt = cn2an.transform(txt, "an2cn")
            uttid2text[cols[0]] = txt
    return uttid2text


def text2tokens(text):
    PUNCTUATIONS = (
        '，。？！,\.?!＂＃＄％＆＇（）＊＋－／：；＜＝＞＠［＼］＾＿｀｛｜｝～｟｠｢｣､　、〃〈〉《》「」『』【】〔〕〖〗〘〙〚〛〜〝〞〟〰〾〿–—‘’‛“”„‟…‧﹏﹑﹔·｡":'
        + "()\[\]{}/;`|=+"
    )
    if text == "":
        return []
    tokens = []

    text = re.sub("<unk>", "", text)
    text = re.sub(r"[%s]+" % PUNCTUATIONS, " ", text)
    text = re.sub("<.*>", "", text)
    text = fix_abbr_simple(text)
    # pattern = re.compile(r'([\u4e00-\u9fff])')
    pattern = re.compile(r"([\u4e00-\u9fff\u3040-\u309f\u30a0-\u30ff\u31f0-\u31ff])")
    parts = pattern.split(text.strip().upper())
    parts = [p for p in parts if len(p.strip()) > 0]
    for part in parts:
        if pattern.fullmatch(part) is not None:
            tokens.append(part)
        else:
            for word in part.strip().split():
                tokens.append(word)
    return tokens


def fix_abbr_simple(text):
    ori_text = text
    # 扔掉超长的
    if re.search(r"(?<![a-zA-Z'])([a-zA-Z] ){5,}[a-zA-Z](?![a-zA-Z'])", text):
        return text

    # 循环处理，直到没有 "X Y" 模式
    prev = None
    while prev != text:
        prev = text
        text = re.sub(
            r"(?<![a-zA-Z'])([a-zA-Z]) (?=[a-zA-Z](?![a-zA-Z']))", r"\1", text
        )

    if ori_text != text:
        print(f"WER FIX: '{ori_text}' --> '{text}'")
    return text


def compute_uttid2wer_info(refs, hyps, print_sentence_wer=False):
    print(f">>> Compute uttid to wer info", flush=True)

    uttid2wer_info = OrderedDict()
    wer_stat = WerStats()
    en_dig_stat = EnDigStats()

    for uttid, ref in refs.items():
        if uttid not in hyps:
            print(f"[WARN] No hyp for {uttid}", flush=True)
            continue
        hyp = hyps[uttid]

        if len(hyp) - len(ref) >= 8:
            print(
                f"[BidLengthDiff]: {uttid} {len(ref)} {len(hyp)}#{' '.join(ref)}#{' '.join(hyp)}"
            )
            # continue

        wer_info = compute_one_wer_info(ref, hyp)
        uttid2wer_info[uttid] = wer_info
        ns = count_english_ditgit(ref, hyp, wer_info)
        wer_stat.add(wer_info)
        en_dig_stat.add(*ns)
        if print_sentence_wer:
            print(f"{uttid} {wer_info}")

    return uttid2wer_info, wer_stat, en_dig_stat


COST_SUB = 3
COST_DEL = 3
COST_INS = 3

ALIGN_CRT = 0
ALIGN_SUB = 1
ALIGN_DEL = 2
ALIGN_INS = 3
ALIGN_END = 4


def compute_one_wer_info(ref, hyp):
    """Impl minimum edit distance and backtrace.
    Args:
        ref, hyp: List[str]
    Returns:
        WerInfo
    """
    ref_len = len(ref)
    hyp_len = len(hyp)

    class _DpPoint:
        def __init__(self, cost, align):
            self.cost = cost
            self.align = align

    dp = []
    for i in range(0, ref_len + 1):
        dp.append([])
        for j in range(0, hyp_len + 1):
            dp[-1].append(_DpPoint(i * j, ALIGN_CRT))

    # Initialize
    for i in range(1, hyp_len + 1):
        dp[0][i].cost = dp[0][i - 1].cost + COST_INS
        dp[0][i].align = ALIGN_INS
    for i in range(1, ref_len + 1):
        dp[i][0].cost = dp[i - 1][0].cost + COST_DEL
        dp[i][0].align = ALIGN_DEL

    # DP
    for i in range(1, ref_len + 1):
        for j in range(1, hyp_len + 1):
            min_cost = 0
            min_align = ALIGN_CRT
            if hyp[j - 1] == ref[i - 1]:
                min_cost = dp[i - 1][j - 1].cost
                min_align = ALIGN_CRT
            else:
                min_cost = dp[i - 1][j - 1].cost + COST_SUB
                min_align = ALIGN_SUB

            del_cost = dp[i - 1][j].cost + COST_DEL
            if del_cost < min_cost:
                min_cost = del_cost
                min_align = ALIGN_DEL

            ins_cost = dp[i][j - 1].cost + COST_INS
            if ins_cost < min_cost:
                min_cost = ins_cost
                min_align = ALIGN_INS

            dp[i][j].cost = min_cost
            dp[i][j].align = min_align

    # Backtrace
    crt = sub = ins = det = 0
    i = ref_len
    j = hyp_len
    align = []
    while i > 0 or j > 0:
        if dp[i][j].align == ALIGN_CRT:
            align.append((i, j, ALIGN_CRT))
            i -= 1
            j -= 1
            crt += 1
        elif dp[i][j].align == ALIGN_SUB:
            align.append((i, j, ALIGN_SUB))
            i -= 1
            j -= 1
            sub += 1
        elif dp[i][j].align == ALIGN_DEL:
            align.append((i, j, ALIGN_DEL))
            i -= 1
            det += 1
        elif dp[i][j].align == ALIGN_INS:
            align.append((i, j, ALIGN_INS))
            j -= 1
            ins += 1

    err = sub + det + ins
    align.reverse()
    wer_info = WerInfo(ref_len, err, crt, sub, det, ins, align)
    return wer_info


class WerInfo:
    def __init__(self, ref, err, crt, sub, dele, ins, ali):
        self.r = ref
        self.e = err
        self.c = crt
        self.s = sub
        self.d = dele
        self.i = ins
        self.ali = ali
        r = max(self.r, 1)
        self.wer = 100.0 * (self.s + self.d + self.i) / r

    def __repr__(self):
        s = f"wer {self.wer:.2f} ref {self.r:2d} sub {self.s:2d} del {self.d:2d} ins {self.i:2d}"
        return s


class WerStats:
    def __init__(self):
        self.infos = []

    def add(self, wer_info):
        self.infos.append(wer_info)

    def print(self):
        r = sum(info.r for info in self.infos)
        if r <= 0:
            print(f"REF len is {r}, check")
            r = 1
        s = sum(info.s for info in self.infos)
        d = sum(info.d for info in self.infos)
        i = sum(info.i for info in self.infos)
        se = 100.0 * s / r
        de = 100.0 * d / r
        ie = 100.0 * i / r
        wer = 100.0 * (s + d + i) / r
        sen = max(len(self.infos), 1)
        errsen = sum(info.e > 0 for info in self.infos)
        ser = 100.0 * errsen / sen
        print("-" * 80)
        print(f"ref{r:6d} sub{s:6d} del{d:6d} ins{i:6d}")
        print(f"WER{wer:6.2f} sub{se:6.2f} del{de:6.2f} ins{ie:6.2f}")
        print(f"SER{ser:6.2f} = {errsen} / {sen}")
        print("-" * 80)


class EnDigStats:
    def __init__(self):
        self.n_en_word = 0
        self.n_en_correct = 0
        self.n_dig_word = 0
        self.n_dig_correct = 0

    def add(self, n_en_word, n_en_correct, n_dig_word, n_dig_correct):
        self.n_en_word += n_en_word
        self.n_en_correct += n_en_correct
        self.n_dig_word += n_dig_word
        self.n_dig_correct += n_dig_correct

    def print(self):
        print(
            f"English #word={self.n_en_word}, #correct={self.n_en_correct}\n"
            f"Digit #word={self.n_dig_word}, #correct={self.n_dig_correct}"
        )
        print("-" * 80)


def count_english_ditgit(ref, hyp, wer_info):
    patt_en = "[a-zA-Z\.\-']+"
    patt_dig = "[0-9]+"
    patt_cjk = re.compile(r"([\u4e00-\u9fff])")
    n_en_word = 0
    n_en_correct = 0
    n_dig_word = 0
    n_dig_correct = 0
    ali = wer_info.ali
    for i, token in enumerate(ref):
        if re.match(patt_en, token):
            n_en_word += 1
            for y in ali:
                if y[0] == i + 1 and y[2] == ALIGN_CRT:
                    j = y[1] - 1
                    n_en_correct += 1
                    break
        if re.match(patt_dig, token):
            n_dig_word += 1
            for y in ali:
                if y[0] == i + 1 and y[2] == ALIGN_CRT:
                    j = y[1] - 1
                    n_dig_correct += 1
                    break
        if (
            not re.match(patt_cjk, token)
            and not re.match(patt_en, token)
            and not re.match(patt_dig, token)
        ):
            print("[WiredChar]:", token)
    return n_en_word, n_en_correct, n_dig_word, n_dig_correct


if __name__ == "__main__":
    args = parser.parse_args()
    print(args, flush=True)
    main(args)



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu)

import logging
import os
import random
import re

import torch
import torch.nn as nn
from transformers import AutoModelForCausalLM

from ..models.fireredasr_aed import FireRedAsrAed
from ..models.module.adapter import Adapter
from ..models.param import count_model_parameters
from ..tokenizer.llm_tokenizer import (
    DEFAULT_SPEECH_TOKEN,
    IGNORE_TOKEN_ID,
    LlmTokenizerWrapper,
)


class FireRedAsrLlm(nn.Module):
    @classmethod
    def load_encoder(cls, model_path):
        assert os.path.exists(model_path)
        package = torch.load(
            model_path, map_location=lambda storage, loc: storage, weights_only=False
        )
        model = FireRedAsrAed.from_args(package["args"])
        if "model_state_dict" in package:
            model.load_state_dict(package["model_state_dict"], strict=False)
        encoder = model.encoder
        encoder_dim = encoder.odim
        return encoder, encoder_dim

    @classmethod
    def from_args(cls, args, freeze_llm=True):
        logging.info(args)
        logging.info("Build FireRedAsrLlm")
        args.freeze_llm = freeze_llm
        # Build Speech Encoder
        encoder, encoder_dim = cls.load_encoder(args.encoder_path)
        count_model_parameters(encoder)
        if args.freeze_encoder:
            logging.info(f"Frezee encoder")
            for name, param in encoder.named_parameters():
                param.requires_grad = False
            encoder.eval()

        # Training use torch.bfloat16
        if args.use_flash_attn:
            attn_implementation = "flash_attention_2"
            if args.use_fp16:
                # torch_dtype = torch.float16
                torch_dtype = torch.bfloat16
            else:
                torch_dtype = torch.float32
        else:
            attn_implementation = "eager"
            if args.use_fp16:
                # torch_dtype = torch.float16
                torch_dtype = torch.bfloat16
            else:
                torch_dtype = torch.float32

        # Build LLM
        llm = AutoModelForCausalLM.from_pretrained(
            args.llm_dir,
            attn_implementation=attn_implementation,
            torch_dtype=torch_dtype,
        )
        count_model_parameters(llm)

        # LLM Freeze or LoRA
        llm_dim = llm.config.hidden_size
        if args.freeze_llm:
            logging.info(f"Frezee LLM")
            for name, param in llm.named_parameters():
                param.requires_grad = False
            llm.eval()
        else:
            if args.use_lora:
                from peft import LoraConfig, get_peft_model

                lora_config = LoraConfig(
                    r=64,
                    lora_alpha=16,
                    target_modules=[
                        "q_proj",
                        "k_proj",
                        "v_proj",
                        "o_proj",
                        "up_proj",
                        "gate_proj",
                        "down_proj",
                    ],
                    lora_dropout=0.05,
                    task_type="CAUSAL_LM",
                )
                llm = get_peft_model(llm, lora_config)
                llm.print_trainable_parameters()

        tokenizer = LlmTokenizerWrapper.build_llm_tokenizer(args.llm_dir)
        assert tokenizer.pad_token_id == tokenizer.convert_tokens_to_ids(
            "<|endoftext|>"
        )
        llm.config.pad_token_id = tokenizer.pad_token_id
        llm.config.bos_token_id = tokenizer.convert_tokens_to_ids("<|im_start|>")
        llm.config.eos_token_id = tokenizer.convert_tokens_to_ids("<|im_end|>")
        llm.config.default_speech_token_id = tokenizer.convert_tokens_to_ids(
            DEFAULT_SPEECH_TOKEN
        )

        # Build projector
        encoder_projector = Adapter(encoder_dim, llm_dim, args.encoder_downsample_rate)
        count_model_parameters(encoder_projector)

        return cls(
            encoder, llm, encoder_projector, args.freeze_encoder, args.freeze_llm
        )

    def __init__(self, encoder, llm, encoder_projector, freeze_encoder, freeze_llm):
        super().__init__()
        self.encoder = encoder
        self.llm = llm
        self.encoder_projector = encoder_projector
        # args
        self.freeze_encoder = freeze_encoder
        self.freeze_llm = freeze_llm
        self.llm_config = llm.config

    def transcribe(
        self,
        padded_feat,
        feat_lengths,
        padded_input_ids,
        attention_mask,
        beam_size=1,
        decode_max_len=0,
        decode_min_len=0,
        repetition_penalty=1.0,
        llm_length_penalty=1.0,
        temperature=1.0,
    ):
        encoder_outs, enc_lengths, enc_mask = self.encoder(padded_feat, feat_lengths)
        speech_features, speech_lens = self.encoder_projector(encoder_outs, enc_lengths)
        inputs_embeds = self.llm.get_input_embeddings()(padded_input_ids)

        inputs_embeds, attention_mask, _ = self._merge_input_ids_with_speech_features(
            speech_features.to(inputs_embeds.dtype),
            inputs_embeds,
            padded_input_ids,
            attention_mask,
            speech_lens=speech_lens,
        )

        max_new_tokens = (
            speech_features.size(1) if decode_max_len < 1 else decode_max_len
        )
        max_new_tokens = max(1, max_new_tokens)

        generated_ids = self.llm.generate(
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            max_new_tokens=max_new_tokens,
            num_beams=beam_size,
            do_sample=False,
            min_length=decode_min_len,
            repetition_penalty=repetition_penalty,
            length_penalty=llm_length_penalty,
            temperature=temperature,
            bos_token_id=self.llm.config.bos_token_id,
            eos_token_id=self.llm.config.eos_token_id,
            pad_token_id=self.llm.config.pad_token_id,
        )

        return generated_ids

    def _merge_input_ids_with_speech_features(
        self,
        speech_features,
        inputs_embeds,
        input_ids,
        attention_mask,
        labels=None,
        speech_lens=None,
    ):
        """
        Modified from: https://github.com/k2-fsa/icefall/blob/master/egs/speech_llm/ASR_LLM/whisper_llm_zh/model.py
        """
        speech_lens = None
        num_speechs, speech_len, embed_dim = speech_features.shape
        batch_size, sequence_length = input_ids.shape
        left_padding = not torch.sum(
            input_ids[:, -1] == torch.tensor(self.llm.config.pad_token_id)
        )
        # 1. Create a mask to know where special speech tokens are
        special_speech_token_mask = input_ids == self.llm.config.default_speech_token_id
        num_special_speech_tokens = torch.sum(special_speech_token_mask, dim=-1)
        # Compute the maximum embed dimension
        max_embed_dim = (
            num_special_speech_tokens.max() * (speech_len - 1)
        ) + sequence_length
        batch_indices, non_speech_indices = torch.where(
            input_ids != self.llm.config.default_speech_token_id
        )

        # 2. Compute the positions where text should be written
        # Calculate new positions for text tokens in merged speech-text sequence.
        # `special_speech_token_mask` identifies speech tokens. Each speech token will be replaced by `nb_text_tokens_per_speechs - 1` text tokens.
        # `torch.cumsum` computes how each speech token shifts subsequent text token positions.
        # - 1 to adjust for zero-based indexing, as `cumsum` inherently increases indices by one.
        new_token_positions = (
            torch.cumsum((special_speech_token_mask * (speech_len - 1) + 1), -1) - 1
        )  # (N,U)
        nb_speech_pad = max_embed_dim - 1 - new_token_positions[:, -1]
        if left_padding:
            new_token_positions += nb_speech_pad[:, None]  # offset for left padding
        text_to_overwrite = new_token_positions[batch_indices, non_speech_indices]

        # 3. Create the full embedding, already padded to the maximum position
        final_embedding = torch.zeros(
            batch_size,
            max_embed_dim,
            embed_dim,
            dtype=inputs_embeds.dtype,
            device=inputs_embeds.device,
        )
        final_attention_mask = torch.zeros(
            batch_size,
            max_embed_dim,
            dtype=attention_mask.dtype,
            device=inputs_embeds.device,
        )
        if labels is not None:
            final_labels = torch.full(
                (batch_size, max_embed_dim),
                IGNORE_TOKEN_ID,
                dtype=input_ids.dtype,
                device=input_ids.device,
            )
        # In case the Vision model or the Language model has been offloaded to CPU, we need to manually
        # set the corresponding tensors into their correct target device.
        target_device = inputs_embeds.device
        batch_indices, non_speech_indices, text_to_overwrite = (
            batch_indices.to(target_device),
            non_speech_indices.to(target_device),
            text_to_overwrite.to(target_device),
        )
        attention_mask = attention_mask.to(target_device)

        # 4. Fill the embeddings based on the mask. If we have ["hey" "<speech>", "how", "are"]
        # we need to index copy on [0, 577, 578, 579] for the text and [1:576] for the speech features
        final_embedding[batch_indices, text_to_overwrite] = inputs_embeds[
            batch_indices, non_speech_indices
        ]
        final_attention_mask[batch_indices, text_to_overwrite] = attention_mask[
            batch_indices, non_speech_indices
        ]
        if labels is not None:
            final_labels[batch_indices, text_to_overwrite] = labels[
                batch_indices, non_speech_indices
            ]

        # 5. Fill the embeddings corresponding to the speechs. Anything that is not `text_positions` needs filling (#29835)
        speech_to_overwrite = torch.full(
            (batch_size, max_embed_dim),
            True,
            dtype=torch.bool,
            device=inputs_embeds.device,
        )
        speech_to_overwrite[batch_indices, text_to_overwrite] = False
        if speech_lens is not None:
            speech_pad_position = speech_to_overwrite.cumsum(-1) <= speech_lens[:, None]
        speech_to_overwrite &= speech_to_overwrite.cumsum(-1) - 1 >= nb_speech_pad[
            :, None
        ].to(target_device)
        if speech_lens is not None:
            if torch.any(speech_lens > speech_len):
                raise ValueError(
                    f"speech_lens contains values ({speech_lens.max()}) larger than "
                    f"speech_len ({speech_len})"
                )

            speech_cumsum = speech_to_overwrite.long().cumsum(-1)
            speech_position_counter = torch.where(
                speech_to_overwrite, speech_cumsum - 1, 0
            )
            valid_speech_positions = speech_position_counter < speech_lens[:, None].to(
                target_device
            )

            speech_to_overwrite &= valid_speech_positions
            if speech_to_overwrite.sum().item() != int(speech_lens.sum().item()):
                raise ValueError(
                    f"speech_lens and speech token distribution mismatch: "
                    f"expected total speech frames {speech_lens.sum().item()}, "
                    f"but got {speech_to_overwrite.sum().item()} positions."
                )
            batch_idx, seq_idx = torch.where(speech_to_overwrite)
            speech_feature_idx = speech_position_counter[speech_to_overwrite]
            final_embedding[batch_idx, seq_idx] = speech_features[
                batch_idx, speech_feature_idx
            ].to(target_device)
        else:
            if speech_to_overwrite.sum() != speech_features.shape[:-1].numel():
                raise ValueError(
                    f"The input provided to the model are wrong. The number of speech tokens is {speech_to_overwrite.sum()} while"
                    f" the number of speech given to the model is {num_speechs}. This prevents correct indexing and breaks batch generation."
                )
            final_embedding[speech_to_overwrite] = (
                speech_features.contiguous()
                .reshape(-1, embed_dim)[: speech_to_overwrite.sum()]
                .to(target_device)
            )

        final_attention_mask[speech_to_overwrite] = 1

        # 6. Mask out the embedding at padding positions, as we later use the past_key_value value to determine the non-attended tokens.
        batch_indices_pad, pad_indices = torch.where(
            input_ids == self.llm.config.pad_token_id
        )
        if len(batch_indices_pad) > 0:
            indices_to_mask = new_token_positions[batch_indices_pad, pad_indices]
            final_embedding[batch_indices_pad, indices_to_mask] = 0
            final_attention_mask[batch_indices_pad, indices_to_mask] = 0

        if labels is None:
            final_labels = None

        return final_embedding, final_attention_mask, final_labels



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Yan Jia)

from .lid import FireRedLid, FireRedLidConfig

__version__ = "0.0.1"

__all__ = [
    "__version__",
    "FireRedLid",
    "FireRedLidConfig",
]



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Kai Huang)

import os
import sys

__version__ = "0.0.1"

_CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

try:
    from fireredasr2s.fireredvad.aed import FireRedAed, FireRedAedConfig
    from fireredasr2s.fireredvad.stream_vad import (
        FireRedStreamVad,
        FireRedStreamVadConfig,
    )
    from fireredasr2s.fireredvad.vad import FireRedVad, FireRedVadConfig
except ImportError:
    if _CURRENT_DIR not in sys.path:
        sys.path.insert(0, _CURRENT_DIR)
    from .aed import FireRedAed, FireRedAedConfig
    from .stream_vad import FireRedStreamVad, FireRedStreamVadConfig
    from .vad import FireRedVad, FireRedVadConfig


def non_stream_vad(wav_path, model_dir="pretrained_models/FireRedVAD/VAD", **kwargs):
    """Quick VAD inference"""
    config = FireRedVadConfig(**kwargs)
    vad = FireRedVad.from_pretrained(model_dir, config)
    result, probs = vad.detect(wav_path)
    return result


def stream_vad_full(
    wav_path, model_dir="pretrained_models/FireRedVAD/Stream-VAD", **kwargs
):
    """Quick Stream VAD inference"""
    config = FireRedStreamVadConfig(**kwargs)
    svad = FireRedStreamVad.from_pretrained(model_dir, config)
    frame_results, result = svad.detect_full(wav_path)
    return frame_results, result


def non_stream_aed(wav_path, model_dir="pretrained_models/FireRedVAD/AED", **kwargs):
    """Quick AED inference"""
    config = FireRedAedConfig(**kwargs)
    aed = FireRedAed.from_pretrained(model_dir, config)
    result, probs = aed.detect(wav_path)
    return result


__all__ = [
    "__version__",
    "FireRedVad",
    "FireRedVadConfig",
    "FireRedAed",
    "FireRedAedConfig",
    "FireRedStreamVad",
    "FireRedStreamVadConfig",
    "non_stream_vad",
    "stream_vad_full",
    "non_stream_aed",
]



from loguru import logger


def print_cuda_info():
    import torch

    # 检查CUDA是否可用
    cuda_available = torch.cuda.is_available()

    if cuda_available:
        # 获取GPU设备数量
        num_gpu = torch.cuda.device_count()

        # 获取当前使用的GPU索引
        current_gpu_index = torch.cuda.current_device()

        # 获取当前GPU的名称
        current_gpu_name = torch.cuda.get_device_name(current_gpu_index)

        # 获取GPU显存的总量和已使用量
        total_memory = torch.cuda.get_device_properties(
            current_gpu_index
        ).total_memory / (1024**3)  # 显存总量(GB)
        used_memory = torch.cuda.memory_allocated(current_gpu_index) / (
            1024**3
        )  # 已使用显存(GB)
        free_memory = total_memory - used_memory  # 剩余显存(GB)

        logger.info(f"CUDA可用，共有 {num_gpu} 个GPU设备可用。")
        logger.info(f"当前使用的GPU设备索引：{current_gpu_index}")
        logger.info(f"当前使用的GPU设备名称：{current_gpu_name}")
        logger.info(f"GPU显存总量：{total_memory:.2f} GB")
        logger.info(f"已使用的GPU显存：{used_memory:.2f} GB")
        logger.info(f"剩余GPU显存：{free_memory:.2f} GB")
    else:
        logger.info("CUDA不可用。")

    # 检查PyTorch版本
    logger.info(f"PyTorch版本：{torch.__version__}")

    import torch

    logger.info(f"CUDA版本：{torch.version.cuda}")



from .detector import VoiceDetector
from .identifier import LanguageIdentifier
from .loader import AudioLoader
from .recogmizer import SpeechRecognizer
from .sentencizer import SpeechSentencizer

__all__ = [
    "AudioLoader",
    "VoiceDetector",
    "LanguageIdentifier",
    "SpeechRecognizer",
    "SpeechSentencizer",
]



from collections import OrderedDict
import io
from pathlib import Path
from queue import Empty, Queue
import threading
import time
from typing import Iterable, List

from loguru import logger
from pydantic import BaseModel, ConfigDict, Field, model_validator, validate_call
from rich.console import Console
from rich.table import Table
from typing_extensions import Self

from .components.base import BaseComponent
from .components.loader import AudioLoader
from .config import Config, registry
from .data import Audio, AudioList
from .utils.base import IOMixin


class AudioQueue(Queue):
    """Create a queue object with a given audio duration limit.

    item in the queue is a tuple of (Audio, Queue).
    """

    def _qsize(self) -> int | bool:
        if len(self.queue) == 0:
            return False  # line 139 Queue._get: while self._qsize() >= self.maxsize: this will be False
        durations = []
        for item in self.queue:
            audio, queue = item
            audio: Audio
            if audio.duration is not None:
                durations.append(audio.duration)
        qsize = round(sum(durations), 2)
        if qsize == 0:
            qsize = 1  # though the audio duration is 0, the queue is not empty, if the queue is empty, will raise Empty
        return qsize


def run_component(component: BaseComponent, audios: AudioList) -> AudioList:
    """Run a component to process the audio data

    Args:
        component (BaseComponent): the component to process the audio data
        audios (AudioList): the audio data to process

    Returns:
        AudioList: the processed audio data
    """
    start = time.perf_counter()
    audios = audios | component
    end = time.perf_counter()
    spent = end - start
    for audio in audios:
        audio: Audio
        if audio.spent_time is None:
            audio.spent_time = {}
        if audio.spent_time.get(component.name) is None:
            audio.spent_time[component.name] = spent
        else:
            audio.spent_time[component.name] += spent
        audio.spent_time[component.name] = round(audio.spent_time[component.name], 3)
    return audios


class PipelineResult(BaseModel):
    model_config: ConfigDict = ConfigDict(arbitrary_types_allowed=True)

    request_queue: AudioQueue | None = Field(None, description="the request queue")
    response_queue: Queue | None = Field(None, description="the response queue")
    loader: BaseComponent | None = Field(None, description="the loader component")

    num_inputs: int = Field(0, description="the number of inputs")
    num_outputs: int = Field(0, description="the number of outputs")

    @validate_call
    def add_task(self, input: str | List[str] | Audio | AudioList):
        """Add a task to the pipeline"""
        input = self.loader._to_audios(input)
        input = run_component(component=self.loader, audios=input)
        self.num_inputs += len(input)
        if isinstance(input, Audio):
            self.request_queue.put((input, self.response_queue))
        elif isinstance(input, AudioList):
            for audio in input:
                self.request_queue.put((audio, self.response_queue))

    def stream_result(self, timeout: float = 100) -> Iterable[Audio]:
        while self.num_outputs < self.num_inputs:
            try:
                audio: Audio = self.response_queue.get(timeout=timeout)
                yield audio
                self.response_queue.task_done()
                self.num_outputs += 1
            except Empty:
                break

    def get_next_result(self, timeout: float = 100) -> Audio | None:
        """Get the next result from the pipeline

        Args:
            timeout (float, optional): the timeout of the response queue waiting for one audio. Defaults to 100. in seconds.

        Returns:
            Audio: the processed audio data.
        """

        try:
            audio: Audio = self.response_queue.get(timeout=timeout)
            self.response_queue.task_done()
            self.num_outputs += 1
            return audio
        except Empty:
            return None


@registry.pipes.register("thread_pipe")
class Pipe(IOMixin):
    """Pipe is a component in the pipeline that can process the audio data in parallel"""

    model_config = ConfigDict(arbitrary_types_allowed=True, validate_assignment=True)

    component: BaseComponent | None = Field(
        None, description="the component to process the audio"
    )
    verbose: bool = Field(False, description="whether to print detailed logs")
    input_queue: AudioQueue | None = None
    output_queue: AudioQueue | None = None
    batch_timeout: float = Field(0.01, description="dynamic batch timeout", ge=0)
    batch_size: int | None = Field(None, description="dynamic batch size", ge=1)
    is_last: bool = Field(True, description="whether it is the last component")
    stop_event: threading.Event | None = Field(None, description="stop event")
    thread: threading.Thread | None = Field(default=None, description="thread object")

    @model_validator(mode="after")
    def validate_stop_event(self):
        if self.stop_event is None:
            self.stop_event = threading.Event()
        return self

    def predict(self, audios: AudioList) -> AudioList:
        """Predict the audio data"""
        audios = run_component(component=self.component, audios=audios)
        return audios

    def run_loop(self):
        """Run the worker loop for this pipeline node.

        The loop keeps running until `stop_event` is set:
        - Pull a dynamic batch from `input_queue` via `dynamic_batch()`.
        - Skip the iteration when no item is collected.
        - Run model/component inference on the batched `AudioList`.
        - Optionally log throughput stats when `verbose` is enabled.

        Output behavior depends on node position:
        - Last node (or no `output_queue`): put each processed audio into its
          per-request `response_queue`.
        - Intermediate node: forward `(audio, response_queue)` tuples to
          `output_queue` for downstream processing.
        """
        while not self.stop_event.is_set():
            batch = self.dynamic_batch()
            if len(batch) == 0:
                continue
            audios = AudioList([audio for audio, _ in batch])
            start = time.perf_counter()
            _ = self.predict(audios)
            end = time.perf_counter()
            spent = round(end - start, 2)
            if self.verbose and spent > 0:
                speed = round(audios.duration_s / spent, 2)
                rtf = (
                    round(spent / audios.duration_s, 3)
                    if audios.duration_s > 0
                    else None
                )

                qsize = self.input_queue.qsize()
                if isinstance(self.input_queue, AudioQueue):
                    # AudioQueue.qsize() represents total queued audio duration (seconds),
                    # and returns False when empty.
                    remaining_s = float(qsize) if qsize is not False else 0.0
                    queue_info = f"remaining={remaining_s}s"
                else:
                    queue_info = f"qsize={qsize} items"
                logger.info(
                    f"{self.component.name} | batch={len(audios)} | audio={round(audios.duration_s, 2)}s | spent={spent}s | speed={speed}x | rtf={rtf} | {queue_info}"
                )
            if self.is_last or self.output_queue is None:
                for audio, response_queue in batch:
                    response_queue: Queue
                    response_queue.put(audio)
            else:
                for item in batch:
                    self.output_queue.put(item)
                    self.input_queue.task_done()

    def dynamic_batch(self):
        """Collect a dynamic batch from the input queue.

        This method tries to accumulate as many items as possible within a
        short time window:
        - Stop when `batch_timeout` is reached.
        - Stop when the number of collected items reaches `batch_size`.
        - If `batch_size` is `None`, it behaves like no count limit.

        Returns:
            list[tuple[Audio, Queue]]: A list of `(audio, response_queue)` pairs.
            It can be empty when no item is available before timeout.
        """
        entered_at = time.monotonic()
        end_time = entered_at + self.batch_timeout
        batch = []
        if self.batch_size is None:
            batch_size = 1e6
        else:
            batch_size = self.batch_size
        while time.monotonic() < end_time and len(batch) < batch_size:
            try:
                audio, response_queue = self.input_queue.get(timeout=self.batch_timeout)
                batch.append((audio, response_queue))
            except Exception:
                break
        return batch

    def start(self):
        self.stop_event.clear()
        self.thread = threading.Thread(target=self.run_loop)
        self.thread.daemon = True
        self.thread.name = self.component.name
        self.thread.start()

    def restart(self):
        """restart the thread"""
        # check if the thread is alive
        if self.thread.is_alive():
            self.stop_event.set()  # set the stop event
            self.thread.join()  # wait for the thread to stop
            # clear output queue
            if self.output_queue:
                while not self.output_queue.empty():
                    self.output_queue.get()
                    self.output_queue.task_done()
            if self.input_queue:
                while not self.input_queue.empty():
                    self.input_queue.get()
                    self.input_queue.task_done()
        self.start()

    def get_config(self) -> Config:
        data = {
            "pipe": {
                "@pipes": "thread_pipe",
                "batch_timeout": self.batch_timeout,
                "batch_size": self.batch_size,
            }
        }
        config = Config(data=data)
        return config

    def save(self, save_dir: str):
        save_dir = Path(save_dir)
        if not save_dir.exists():
            save_dir.mkdir(parents=True, exist_ok=True)
        config = self.get_config()
        config.to_disk(save_dir / "config.cfg")
        self.component.save(save_dir / "component")

    def load(self, save_dir: str):
        save_dir = Path(save_dir)
        config = Config().from_disk(save_dir / "config.cfg")
        pipe = registry.resolve(config)["pipe"]
        self.batch_size = pipe.batch_size
        self.batch_timeout = pipe.batch_timeout
        component_dir = save_dir / "component"
        component_config = Config().from_disk(component_dir / "config.cfg")
        component: BaseComponent = registry.resolve(component_config)["component"]
        self.component = component.load(component_dir)
        return self

    @property
    def name(self) -> str | None:
        if self.component is None:
            return None
        return self.component.name

    @property
    def input_tags(self) -> List[str]:
        if self.component is None:
            return []
        return self.component.input_tags

    @property
    def output_tags(self) -> List[str]:
        if self.component is None:
            return []
        return self.component.output_tags


class AudioPipeline(IOMixin):
    """Audio Pipeline start with a loader component which can load the audio data from the disk or other sources.
    Args:
        loader (BaseComponent, optional): the loader component. Defaults to AudioLoader.
        pipes (OrderedDict[str, Pipe] | None, optional): the components in the pipeline. Defaults to None.
        request_queue (AudioQueue | None, optional): the request queue. Defaults to None.
    """

    model_config: ConfigDict = ConfigDict(arbitrary_types_allowed=True)

    loader: BaseComponent = Field(
        default_factory=AudioLoader, description="the loader component"
    )
    pipes: OrderedDict[str, Pipe] | None = Field(
        default=None, description="the components in the pipeline"
    )
    request_queue: AudioQueue | None = Field(
        default=None, description="the request queue"
    )

    @model_validator(mode="after")
    def validate_pipes(self):
        if self.pipes is None:
            self.pipes = OrderedDict()
        if self.request_queue is None:
            self.request_queue = AudioQueue()
        if isinstance(self.loader, str):
            self.loader = registry.components.get(self.loader)()
        return self

    @validate_call
    def __call__(
        self,
        input: str | Audio | Path,
        verbose: bool = False,
        clear: bool = False,
        timeout: float | None = None,
    ) -> Audio:
        return self.run(input, verbose=verbose, clear=clear, timeout=timeout)[0]

    @validate_call
    def run(
        self,
        input: str | Path | Audio | List[str] | AudioList | List[Path],
        timeout: float | None = None,
        verbose: bool = False,
        clear: bool = False,
    ) -> AudioList:
        """Run the pipeline

        Args:
            input (str | Audio | List[str] | AudioList): the input data, can be a path, an Audio object, a list of paths or a list of Audio objects.
            timeout (float | None, optional): the timeout of the response queue. Defaults to None.
            verbose (bool, optional): whether to print detailed logs. Defaults to False.
            clear (bool, optional): whether to clear the audio data after processing. Defaults to False.

        Returns:
            AudioList: the processed audio data
        """
        for _, pipe in self.pipes.items():
            pipe.verbose = verbose
        response_queue = Queue()
        audios = self.loader._to_audios(input)
        for audio in audios:
            if audio.is_bad:
                response_queue.put(audio)
        results = AudioList()
        start = time.perf_counter()
        for i in range(0, len(audios), 10):
            batch_audios = audios[i : i + 10]
            batch_audios = run_component(component=self.loader, audios=batch_audios)
            if len(self.pipe_names) > 0:
                for audio in batch_audios:
                    self.request_queue.put((audio, response_queue))
            else:
                results.extend(batch_audios)

        while True:
            if len(audios) == len(results):
                break
            try:
                audio: Audio = response_queue.get(timeout=timeout)
                if clear:
                    audio.clear()
                results.append(audio)
                response_queue.task_done()
            except Empty:
                logger.error("response queue timeout")
                for audio in audios:
                    if not audio.is_bad:
                        audio.is_bad = True
                        audio.bad_reason = "response queue timeout"
                break
            except KeyboardInterrupt:
                logger.error("KeyboardInterrupt, waiting for restart")
                self.restart()
                logger.info("Pipeline restart success")
                break
        end = time.perf_counter()
        spent = max(round(end - start, 2), 1e-5)
        if len(audios) == len(results):
            if not response_queue.empty():
                logger.error("response not equal to request")
                for audio in audios:
                    if not audio.is_bad:
                        audio.is_bad = True
                        audio.bad_reason = "response not equal to request"
            else:
                if verbose:
                    if (
                        not audios.has_bad_audio()
                        and spent > 0
                        and len(self.pipe_names) > 0
                    ):
                        speed = round(results.duration_s / spent, 2)
                        rtf = (
                            round(spent / results.duration_s, 3)
                            if results.duration_s > 0
                            else None
                        )
                        logger.opt(colors=True).info(
                            "<bold>pipeline</bold> | "
                            f"<cyan>pipes</cyan>={len(self.pipe_names)} | "
                            f"<cyan>batch</cyan>={len(results)} | "
                            f"<cyan>audio</cyan>={results.duration_s:0.2f}s | "
                            f"<cyan>spent</cyan>={spent:0.2f}s | "
                            f"<green>speed</green>=<green>{speed}x</green> | "
                            f"<magenta>rtf</magenta>=<magenta>{rtf}</magenta>"
                        )
        return audios

    @validate_call
    def stream(
        self,
        input: str | Audio | List[str] | AudioList,
        timeout: float | None = None,
        clear: bool = False,
        verbose: bool = False,
    ) -> Iterable[Audio]:
        """Stream the audio data. return the processed audio data one by one.
        Args:
            input (str | Audio | List[str] | AudioList): the input data, can be a path, an Audio object, a list of paths or a list of Audio objects.
            timeout (float | None, optional): the timeout of the response queue waiting for one audio. Defaults to None.
            clear (bool, optional): whether to clear the audio data after processing. Defaults to True.
            verbose (bool, optional): whether to print detailed logs. Defaults to False.

        Yields:
            Iterable[Audio]: processed audio data
        """
        for _, pipe in self.pipes.items():
            pipe.verbose = verbose
        response_queue = Queue()
        audios = self.loader._to_audios(input)
        results = AudioList()
        for i in range(0, len(audios), 10):
            batch_audios = audios[i : i + 10]
            batch_audios = run_component(component=self.loader, audios=batch_audios)
            if len(self.pipe_names) > 0:
                for audio in batch_audios:
                    self.request_queue.put((audio, response_queue))
            else:
                results.extend(batch_audios)
        if len(self.pipe_names) == 0:
            yield from audios
            return

        start = time.perf_counter()
        while True:
            if len(audios) == len(results):
                break
            try:
                audio: Audio = response_queue.get(timeout=timeout)
                if clear:
                    audio.clear()
                yield audio
                results.append(audio)
                response_queue.task_done()
            except Empty:
                logger.error("response queue timeout")
                bad_audios = audios - results
                for audio in bad_audios:
                    if not audio.is_bad:
                        audio.is_bad = True
                        audio.bad_reason = "response queue timeout"
                yield from bad_audios
                break
        end = time.perf_counter()
        spent = max(round(end - start, 2), 1e-5)
        if verbose:
            if not audios.has_bad_audio() and spent > 0 and len(self.pipe_names) > 0:
                logger.info(
                    f"pipeline processed {results.duration_s:0.2f}s {len(results)} audios, spent {spent:0.2f}s, speed {round(results.duration_s / spent, 2)}x"
                )

    @validate_call
    def infinite(
        self, input: str | Audio | List[str] | AudioList, verbose: bool = False
    ) -> PipelineResult:
        """run the pipeline in infinite mode which can add tasks and stream results"""
        for _, pipe in self.pipes.items():
            pipe.verbose = verbose
        audios = self.loader._to_audios(input)
        response_queue = Queue()
        result = PipelineResult(
            request_queue=self.request_queue,
            response_queue=response_queue,
            loader=self.loader,
        )
        result.num_inputs = len(audios)
        result.num_outputs = 0
        for i in range(0, len(audios), 10):
            batch_audios = audios[i : i + 10]
            batch_audios = run_component(component=self.loader, audios=batch_audios)
            if len(self.pipe_names) > 0:
                for audio in batch_audios:
                    self.request_queue.put((audio, response_queue))
            else:
                result.num_outputs += len(batch_audios)
        return result

    @validate_call
    def add_pipe(
        self,
        component: str,
        model: str,
        batch_timeout: float = 0.1,
        batch_size: int | None = None,
        queue_duration: int = -1,
        verbose: bool = False,
        **config,
    ) -> Self:
        """Add a component to the pipeline
        Args:
            component (str): the component to add to the pipeline.
            model (str): the model name of the component.
            batch_timeout (float, optional): dynamic batch timeout. Defaults to 0.1.
            batch_size (int | None, optional): dynamic batch size. Defaults to None.
            queue_duration (int, optional): duration of the queue. Defaults to -1. If -1, the queue duration is infinite. if the queue's duration is above this value, the previous component will be blocked.
            verbose (bool, optional): whether to print detailed logs. Defaults to False.
            **config: the parameters of the component.
        """
        if component == "loader":
            raise ValueError("loader component is reserved")
        if len(self.pipes) == 0:
            input_queue = self.request_queue
        else:
            input_queue = AudioQueue(maxsize=queue_duration)
            pre_pipe = self.pipes[self.pipe_names[-1]]
            pre_pipe.output_queue = input_queue
            pre_pipe.is_last = False
        pipe: Pipe = self._init_pipe(
            component=component,
            model=model,
            is_last=True,
            batch_timeout=batch_timeout,
            batch_size=batch_size,
            input_queue=input_queue,
            verbose=verbose,
            **config,
        )
        pipe.start()
        self.pipes[pipe.name] = pipe
        return self

    def get_pipe(self, name: str) -> Pipe | None:
        """Get a component from the pipeline

        Args:
            name (str): the name of the component registered in the registry.

        Returns:
            Pipe | None: the component object in the pipeline or None
        """
        return self.pipes.get(name, None)

    def _init_pipe(
        self,
        component: str | BaseComponent,
        model: str,
        verbose: bool = False,
        is_last: bool = True,
        input_queue: AudioQueue = None,
        output_queue: AudioQueue = None,
        batch_timeout: float = 0.01,
        batch_size: int | None = None,
        **config,
    ) -> "Pipe":
        """Initialize a component in the pipeline

        Args:
            component (str | BaseComponent): the component to add to the pipeline. It can be the name of the component registered in the registry or the component object.
            verbose (bool, optional): whether to print detailed logs. Defaults to False.
            is_last (bool, optional): whether it is the last component. Defaults to True.
            input_queue (AudioQueue, optional): the input queue of the component. Defaults to None.
            output_queue (AudioQueue, optional): the output queue of the component. Defaults to None.
            batch_timeout (float, optional): dynamic batch timeout. Defaults to 0.01.
            batch_size (int | None, optional): dynamic batch size. Defaults to None.

        Returns:
            Pipe: the component object in the pipeline
        """
        if isinstance(component, str):
            _component: BaseComponent = registry.components.get(component)()
            _component = _component.setup(model=model, **config)
            pipe = Pipe(
                component=_component,
                verbose=verbose,
                is_last=is_last,
                input_queue=input_queue,
                output_queue=output_queue,
                batch_timeout=batch_timeout,
                batch_size=batch_size,
            )
            return pipe
        elif isinstance(component, BaseComponent):
            pipe = Pipe(
                component=component,
                verbose=verbose,
                is_last=is_last,
                input_queue=input_queue,
                output_queue=output_queue,
                batch_timeout=batch_timeout,
                batch_size=batch_size,
            )
            return pipe

    def remove_pipe(self, name: str) -> Self:
        """Remove a component from the pipeline

        Args:
            name (str): the name of the component.

        Returns:
            Pipeline: the pipeline object
        """
        pipe: Pipe = self.pipes[name]
        if len(self.pipes) == 1:
            self._del_pipe(name)
            return self
        elif pipe.is_last:
            pre_pipe = self.pipes[self.pipe_names[-2]]
            pre_pipe.is_last = True
            pre_pipe.output_queue = None
            self._del_pipe(name)
            return self
        else:
            for i, pipe_name in enumerate(self.pipe_names):
                if pipe_name == name:
                    if i == 0:
                        next_pipe = self.pipes[self.pipe_names[i + 1]]
                        next_pipe.input_queue = self.request_queue
                        self._del_pipe(name)
                        return self
                    else:
                        pre_pipe = self.pipes[self.pipe_names[i - 1]]
                        next_pipe = self.pipes[self.pipe_names[i + 1]]
                        next_pipe.input_queue = pre_pipe.output_queue
                        self._del_pipe(name)
                        return self

    def set_pipe(
        self,
        name: str,
        batch_size: int | None = None,
        batch_timeout: float = 0.1,
        **config,
    ) -> Self:
        """Set the dynamic batch size and batch timeout of a component

        Args:
            name (str): the name of the pipe
            batch_size (int | None, optional): dynamic batch size. Defaults to None.
            batch_timeout (float, optional): dynamic batch timeout. Defaults to 0.1.
            **config: the parameters of the component.

        Returns:
            Pipeline: the pipeline object
        """
        self.pipes[name].batch_size = batch_size
        self.pipes[name].batch_timeout = batch_timeout
        if config:
            for key, value in config.items():
                if hasattr(self.pipes[name].component, key):
                    setattr(self.pipes[name].component, key, value)
        return self

    def set_loader(self, **kwargs) -> Self:
        """Set the parameters of the loader component

        Returns:
            Self: the pipeline object
        """
        if kwargs:
            for key, value in kwargs.items():
                if hasattr(self.loader, key):
                    setattr(self.loader, key, value)
                else:
                    logger.warning(f"{key} not found in loader")
        return self

    def get_config(self) -> Config:
        """Get the configuration of the pipeline

        Returns:
            Config: the configuration of the pipeline
        """
        data = {
            "pipeline": {
                "@pipelines": "audio_pipeline",
                "loader": self.loader.get_config()["component"],
                "pipes": {},
            }
        }
        for name, pipe in self.pipes.items():
            data["pipeline"]["pipes"][name] = pipe.get_config()["pipe"]
        config = Config(data=data)
        return config

    def save(self, save_dir: str):
        save_dir = Path(save_dir)
        if not save_dir.exists():
            save_dir.mkdir(parents=True, exist_ok=True)
        else:
            import shutil

            shutil.rmtree(save_dir)
            save_dir.mkdir(parents=True, exist_ok=True)
        config = self.get_config()
        config.to_disk(save_dir / "config.cfg")
        try:
            for name, pipe in self.pipes.items():
                pipe.save(save_dir=save_dir / name)
        except Exception as e:
            import shutil

            shutil.rmtree(save_dir)
            raise e

    def load(self, save_dir: str) -> Self:
        save_dir = Path(save_dir)
        if not save_dir.exists():
            raise FileNotFoundError(f"{save_dir} not found")
        config = Config().from_disk(save_dir / "config.cfg")
        pipeline = registry.resolve(config)["pipeline"]
        self.loader = pipeline.loader
        for name, pipe in pipeline.pipes.items():
            pipe: Pipe = pipe.load(save_dir / name)
            self.add_pipe(
                component=pipe.component,
                batch_timeout=pipe.batch_timeout,
                batch_size=pipe.batch_size,
            )
        return self

    def restart(self):
        while not self.request_queue.empty():
            self.request_queue.get()
            self.request_queue.task_done()
        for _, pipe in self.pipes.items():
            pipe.restart()
        return self

    @property
    def pipe_names(self) -> List[Pipe]:
        """Get the names of the components in the pipeline"""
        names = []
        for _, pipe in self.pipes.items():
            names.append(pipe.name)
        return names

    @property
    def pipe_registry_names(self) -> List[str]:
        """Get the names of the components in the pipeline from the registry

        Returns:
            List[str]: the names of the components in the pipeline
        """
        names = list(self.pipes.keys())
        return names

    def _del_pipe(self, name: str):
        """delete a component from the pipeline"""
        pipe: Pipe = self.pipes[name]
        pipe.stop_event.set()
        pipe.thread.join()
        del self.pipes[name]

    def __del__(self):
        """terminate all threads when delete the object"""
        for name in self.pipe_names:
            pipe = self.pipes[name]
            if pipe.stop_event is not None:
                pipe.stop_event.set()
            if pipe.thread is not None:
                pipe.thread.join()

    def _build_rich_table(self) -> Table:
        table = Table(
            title="AudioPipeline",
            title_style="bold magenta",
            show_lines=True,
            expand=False,
        )
        table.add_column("#", style="dim", justify="right", width=4)
        table.add_column("Component", style="cyan bold")
        table.add_column("Type", style="green")
        table.add_column("Input Tags", style="yellow")
        table.add_column("Output Tags", style="yellow")

        components = [("loader", self.loader)] + [
            (name, self.pipes[name].component) for name in self.pipe_names
        ]
        for i, (name, comp) in enumerate(components):
            comp_type = type(comp).__name__ if comp else "None"
            in_tags = ", ".join(comp.input_tags) if comp and comp.input_tags else "-"
            out_tags = ", ".join(comp.output_tags) if comp and comp.output_tags else "-"
            table.add_row(str(i), name, comp_type, in_tags, out_tags)

        return table

    def __str__(self) -> str:
        buf = io.StringIO()
        console = Console(file=buf, force_terminal=True, width=120)
        console.print(self._build_rich_table())
        return buf.getvalue().rstrip()

    def __repr__(self) -> str:
        return self.__str__()

    def __len__(self) -> int:
        return len(self.pipes)



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Kai Huang)

from dataclasses import dataclass
import logging
import os

import torch

from .core.audio_feat import AudioFeat
from .core.detect_model import DetectModel
from .core.vad_postprocessor import VadPostprocessor

logger = logging.getLogger(__name__)


@dataclass
class FireRedAedConfig:
    use_gpu: bool = False
    smooth_window_size: int = 5
    speech_threshold: float = 0.4
    singing_threshold: float = 0.5
    music_threshold: float = 0.5
    min_event_frame: int = 20
    max_event_frame: int = 2000  # 20s
    min_silence_frame: int = 20
    merge_silence_frame: int = 0
    extend_speech_frame: int = 0
    chunk_max_frame: int = 30000  # 300s


class FireRedAed:
    IDX2EVENT = {0: "speech", 1: "singing", 2: "music"}

    @classmethod
    def from_pretrained(cls, model_dir, config=FireRedAedConfig()):
        # Build Feat Extractor
        cmvn_path = os.path.join(model_dir, "cmvn.ark")
        audio_feat = AudioFeat(cmvn_path)

        # Build Model
        model = DetectModel.from_pretrained(model_dir)
        if config.use_gpu:
            model.cuda()
        else:
            model.cpu()

        # Build Postprocessor
        event2postprocessor = {}
        for event in cls.IDX2EVENT.values():
            threshold = getattr(config, f"{event}_threshold")
            event2postprocessor[event] = VadPostprocessor(
                config.smooth_window_size,
                threshold,
                config.min_event_frame,
                config.max_event_frame,
                config.min_silence_frame,
                config.merge_silence_frame,
                config.extend_speech_frame,
            )
        return cls(audio_feat, model, event2postprocessor, config)

    def __init__(self, audio_feat, model, event2postprocessor, config):
        self.audio_feat = audio_feat
        self.model = model
        self.event2postprocessor = event2postprocessor
        self.config = config

    def detect(self, audio):
        # Extract feat
        feat, dur = self.audio_feat.extract(audio)
        if self.config.use_gpu:
            feat = feat.cuda()

        # Model inference
        if feat.size(0) <= self.config.chunk_max_frame:
            probs, _ = self.model.forward(feat.unsqueeze(0))
            assert probs.size(-1) == len(self.IDX2EVENT)
            probs = probs.cpu().squeeze(0)  # (T,3)
        else:
            logger.warning(
                f"Too long input, split every {self.config.chunk_max_frame} frames"
            )
            chunk_probs = []
            chunks = feat.split(self.config.chunk_max_frame, dim=0)
            for chunk in chunks:
                chunk_prob, _ = self.model.forward(chunk.unsqueeze(0))
                assert chunk_prob.size(-1) == len(self.IDX2EVENT)
                chunk_probs.append(chunk_prob.cpu())
            probs = torch.cat(chunk_probs, dim=1)
            probs = probs.squeeze(0)  # (T,3)

        # Prob Postprocess
        event2starts_ends_s = {}
        event2raw_ratio = {}
        for idx, event in self.IDX2EVENT.items():
            threshold = getattr(self.config, f"{event}_threshold")
            postprocessor = self.event2postprocessor[event]
            event_probs = probs[:, idx].tolist()
            decision = postprocessor.process(event_probs)
            starts_ends_s = postprocessor.decision_to_segment(decision, dur)
            event2starts_ends_s[event] = starts_ends_s

            raw_ratio = (
                sum(int(p >= threshold) for p in event_probs) / len(event_probs)
                if len(event_probs)
                else 0
            )
            event2raw_ratio[event] = round(raw_ratio, 3)

        # Format result
        result = {
            "dur": round(dur, 3),
            "event2timestamps": event2starts_ends_s,
            "event2ratio": event2raw_ratio,
        }
        if isinstance(audio, str):
            result["wav_path"] = audio
        return result, probs



from typing import List

from joblib import Parallel, delayed
from typing_extensions import Self

from fasr.config import registry
from fasr.data.audio import (
    AudioChannel,
    AudioList,
)
from fasr.model import PuncModel

from .base import BaseComponent


@registry.components.register("sentencizer")
class SpeechSentencizer(BaseComponent):
    """句子分割器"""

    name: str = "sentencizer"
    description: str = "Sentence segments from audio channels"
    input_tags: List[str] = ["Audio.channels"]
    output_tags: List[str] = ["AudioChannel.sents"]
    model: PuncModel | None = None
    num_threads: int = 1

    def predict(self, audios: AudioList) -> AudioList:
        channels = []
        for audio in audios:
            if audio.channels is not None:
                channels.extend(audio.channels)
        _ = Parallel(n_jobs=self.num_threads, prefer="threads")(
            delayed(self.predict_channel)(channel) for channel in channels
        )
        return audios

    def predict_channel(self, channel: AudioChannel) -> AudioChannel:
        text = channel.text
        if text.strip() != "":
            sents = self.model.restore(text=text)
            channel.sents = sents
            sent_tokens = []
            seg_tokens = []
            if channel.segments is not None:
                for seg in channel.segments:
                    seg_tokens.extend(seg.tokens)
            for sent in channel.sents:
                sent_tokens.extend(sent.tokens)
            if len(seg_tokens) == len(sent_tokens):
                for sent_token, seg_token in zip(sent_tokens, seg_tokens):
                    if sent_token.text == seg_token.text:
                        sent_token.start_ms = seg_token.start_ms
                        sent_token.end_ms = seg_token.end_ms
            for sent in sents:
                sent.start_ms = sent.tokens[0].start_ms
                sent.end_ms = sent.tokens[-1].end_ms
        return channel

    def setup(
        self,
        model: str,
        num_threads: int = 1,
        **kwargs,
    ) -> Self:
        self.load_model_plugin(group="fasr.punc_models", model=model)
        model: PuncModel = registry.punc_models.get(model)()
        self.model = model.from_checkpoint(**kwargs)
        self.num_threads = num_threads
        return self



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu)

import torch
import torch.nn as nn


class Adapter(nn.Module):
    def __init__(self, encoder_dim, llm_dim, downsample_rate=2):
        super().__init__()
        self.ds = downsample_rate
        self.linear1 = nn.Linear(encoder_dim * downsample_rate, llm_dim)
        self.relu = nn.ReLU()
        self.linear2 = nn.Linear(llm_dim, llm_dim)

    def forward(self, x, x_lens):
        batch_size, seq_len, feat_dim = x.size()
        num_frames_to_discard = seq_len % self.ds
        if num_frames_to_discard > 0:
            x = x[:, :-num_frames_to_discard, :]
        seq_len = x.size(1)

        x = x.contiguous()
        x = x.view(batch_size, seq_len // self.ds, feat_dim * self.ds)

        x = self.linear1(x)
        x = self.relu(x)
        x = self.linear2(x)

        new_x_lens = torch.clamp(x_lens, max=seq_len) // self.ds
        return x, new_x_lens



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Yan Jia)

from typing import Dict, List, Optional

import torch
from torch import Tensor
import torch.nn as nn
import torch.nn.functional as F


class TransformerDecoder(nn.Module):
    def __init__(
        self,
        sos_id,
        eos_id,
        pad_id,
        odim,
        n_layers,
        n_head,
        d_model,
        residual_dropout=0.1,
        pe_maxlen=5000,
    ):
        super().__init__()
        self.INF = 1e10
        self.pad_id = pad_id
        self.sos_id = sos_id
        self.eos_id = eos_id
        self.n_layers = n_layers

        self.tgt_word_emb = nn.Embedding(odim, d_model, padding_idx=self.pad_id)
        self.positional_encoding = PositionalEncoding(d_model, max_len=pe_maxlen)
        self.dropout = nn.Dropout(residual_dropout)

        self.layer_stack = nn.ModuleList()
        for l in range(n_layers):
            block = DecoderLayer(d_model, n_head, residual_dropout)
            self.layer_stack.append(block)

        self.tgt_word_prj = nn.Linear(d_model, odim, bias=False)
        self.layer_norm_out = nn.LayerNorm(d_model)

        self.tgt_word_prj.weight = self.tgt_word_emb.weight
        self.scale = d_model**0.5

    def batch_beam_search(
        self,
        encoder_outputs,
        src_masks,
        beam_size=1,
        nbest=1,
        decode_max_len=0,
        softmax_smoothing=1.0,
        length_penalty=0.0,
        eos_penalty=1.0,
    ):
        B = beam_size
        N, Ti, H = encoder_outputs.size()
        device = encoder_outputs.device
        maxlen = decode_max_len if decode_max_len > 0 else Ti
        assert eos_penalty > 0.0

        encoder_outputs = (
            encoder_outputs.unsqueeze(1).repeat(1, B, 1, 1).view(N * B, Ti, H)
        )
        src_mask = src_masks.unsqueeze(1).repeat(1, B, 1, 1).view(N * B, -1, Ti)
        ys = torch.ones(N * B, 1).fill_(self.sos_id).long().to(device)
        t_ys = ys.clone()
        confidences = torch.zeros(N * B, 1).float().to(device)
        caches: List[Optional[Tensor]] = []
        for _ in range(self.n_layers):
            caches.append(None)
        scores = torch.tensor([0.0] + [-self.INF] * (B - 1)).float().to(device)
        scores = scores.repeat(N).view(N * B, 1)
        is_finished = torch.zeros_like(scores)

        for t in range(maxlen):
            tgt_mask = self.ignored_target_position_is_0(ys, self.pad_id)

            dec_output = self.dropout(
                self.tgt_word_emb(ys) * self.scale + self.positional_encoding(ys)
            )
            i = 0
            for dec_layer in self.layer_stack:
                dec_output = dec_layer.forward(
                    dec_output,
                    encoder_outputs,
                    tgt_mask,
                    src_mask,
                    cache=caches[i],
                )
                caches[i] = dec_output
                i += 1

            dec_output = self.layer_norm_out(dec_output)

            t_logit = self.tgt_word_prj(dec_output[:, -1])
            t_scores = F.log_softmax(t_logit / softmax_smoothing, dim=-1)
            t_origin_scores = t_scores  # noqa: F841

            if eos_penalty != 1.0:
                t_scores[:, self.eos_id] *= eos_penalty

            t_topB_scores, t_topB_ys = torch.topk(t_scores, k=B, dim=1)
            t_topB_scores = self.set_finished_beam_score_to_zero(
                t_topB_scores, is_finished
            )
            t_topB_ys = self.set_finished_beam_y_to_eos(t_topB_ys, is_finished)

            scores = scores + t_topB_scores

            scores = scores.view(N, B * B)
            scores, topB_score_ids = torch.topk(scores, k=B, dim=1)
            scores = scores.view(-1, 1)

            topB_row_number_in_each_B_rows_of_ys = torch.div(topB_score_ids, B).view(
                N * B
            )
            stride = B * torch.arange(N).view(N, 1).repeat(1, B).view(N * B).to(device)
            topB_row_number_in_ys = (
                topB_row_number_in_each_B_rows_of_ys.long() + stride.long()
            )

            ys = ys[topB_row_number_in_ys]
            t_ys = torch.gather(
                t_topB_ys.view(N, B * B), dim=1, index=topB_score_ids
            ).view(N * B, 1)
            ys = torch.cat((ys, t_ys), dim=1)

            confidences = confidences[topB_row_number_in_ys]
            t_confidences = torch.gather(
                t_topB_scores.view(N, B * B), dim=1, index=topB_score_ids
            ).view(N * B, 1)
            t_confidences = torch.exp(t_confidences)
            assert torch.all(t_confidences <= 1.0)
            assert torch.all(t_confidences >= 0.0)
            confidences = torch.cat((confidences, t_confidences), dim=1)

            new_caches: List[Optional[Tensor]] = []
            for cache in caches:
                if cache is not None:
                    new_caches.append(cache[topB_row_number_in_ys])
            caches = new_caches

            is_finished = t_ys.eq(self.eos_id)
            if is_finished.sum().item() == N * B:
                break

        scores = scores.view(N, B)
        ys = ys.view(N, B, -1)
        ys_lengths = self.get_ys_lengths(ys)
        if length_penalty > 0.0:
            penalty = torch.pow((5 + ys_lengths.float()) / (5.0 + 1), length_penalty)
            scores /= penalty
        nbest_scores, nbest_ids = torch.topk(scores, k=int(nbest), dim=1)
        nbest_scores = -1.0 * nbest_scores
        index = nbest_ids + B * torch.arange(N).view(N, 1).to(device).long()
        nbest_ys = ys.view(N * B, -1)[index.view(-1)]
        nbest_ys = nbest_ys.view(N, nbest_ids.size(1), -1)
        nbest_ys_lengths = ys_lengths.view(N * B)[index.view(-1)].view(N, -1)
        nbest_confidences = confidences.view(N * B, -1)[index.view(-1)].view(
            N, nbest_ids.size(1), -1
        )

        nbest_hyps: List[List[Dict[str, Tensor]]] = []
        for n in range(N):
            n_nbest_hyps: List[Dict[str, Tensor]] = []
            for i, score in enumerate(nbest_scores[n]):
                confidence = nbest_confidences[n, i, 1 : nbest_ys_lengths[n, i]]
                confidence = confidence.mean()
                new_hyp = {
                    "yseq": nbest_ys[n, i, 1 : nbest_ys_lengths[n, i]],
                    "confidence": confidence,
                }
                n_nbest_hyps.append(new_hyp)
            nbest_hyps.append(n_nbest_hyps)
        return nbest_hyps

    def ignored_target_position_is_0(self, padded_targets, ignore_id):
        mask = torch.ne(padded_targets, ignore_id)
        mask = mask.unsqueeze(dim=1)
        T = padded_targets.size(-1)
        upper_tri_0_mask = self.upper_triangular_is_0(T).unsqueeze(0).to(mask.dtype)
        upper_tri_0_mask = upper_tri_0_mask.to(mask.dtype).to(mask.device)
        return mask.to(torch.uint8) & upper_tri_0_mask.to(torch.uint8)

    def upper_triangular_is_0(self, size):
        ones = torch.ones(size, size)
        tri_left_ones = torch.tril(ones)
        return tri_left_ones.to(torch.uint8)

    def set_finished_beam_score_to_zero(self, scores, is_finished):
        NB, B = scores.size()
        is_finished = is_finished.float()
        mask_score = (
            torch.tensor([0.0] + [-self.INF] * (B - 1)).float().to(scores.device)
        )
        mask_score = mask_score.view(1, B).repeat(NB, 1)
        return scores * (1 - is_finished) + mask_score * is_finished

    def set_finished_beam_y_to_eos(self, ys, is_finished):
        is_finished = is_finished.long()
        return ys * (1 - is_finished) + self.eos_id * is_finished

    def get_ys_lengths(self, ys):
        N, B, Tmax = ys.size()
        ys_lengths = torch.sum(torch.ne(ys, self.eos_id), dim=-1)
        return ys_lengths.int()


class DecoderLayer(nn.Module):
    def __init__(self, d_model, n_head, dropout):
        super().__init__()
        self.self_attn_norm = nn.LayerNorm(d_model)
        self.self_attn = DecoderMultiHeadAttention(d_model, n_head, dropout)

        self.cross_attn_norm = nn.LayerNorm(d_model)
        self.cross_attn = DecoderMultiHeadAttention(d_model, n_head, dropout)

        self.mlp_norm = nn.LayerNorm(d_model)
        self.mlp = PositionwiseFeedForward(d_model, d_model * 4, dropout)

    def forward(
        self,
        dec_input,
        enc_output,
        self_attn_mask,
        cross_attn_mask,
        cache=None,
    ):
        x = dec_input
        residual = x
        x = self.self_attn_norm(x)
        if cache is not None:
            xq = x[:, -1:, :]
            residual = residual[:, -1:, :]
            self_attn_mask = self_attn_mask[:, -1:, :]
        else:
            xq = x
        x = self.self_attn(xq, x, x, mask=self_attn_mask)
        x = residual + x

        residual = x
        x = self.cross_attn_norm(x)
        x = self.cross_attn(x, enc_output, enc_output, mask=cross_attn_mask)
        x = residual + x

        residual = x
        x = self.mlp_norm(x)
        x = residual + self.mlp(x)

        if cache is not None:
            x = torch.cat([cache, x], dim=1)

        return x


class DecoderMultiHeadAttention(nn.Module):
    def __init__(self, d_model, n_head, dropout=0.1):
        super().__init__()
        self.d_model = d_model
        self.n_head = n_head
        self.d_k = d_model // n_head

        self.w_qs = nn.Linear(d_model, n_head * self.d_k)
        self.w_ks = nn.Linear(d_model, n_head * self.d_k, bias=False)
        self.w_vs = nn.Linear(d_model, n_head * self.d_k)

        self.attention = DecoderScaledDotProductAttention(temperature=self.d_k**0.5)
        self.fc = nn.Linear(n_head * self.d_k, d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, q, k, v, mask=None):
        bs = q.size(0)

        q = self.w_qs(q).view(bs, -1, self.n_head, self.d_k)
        k = self.w_ks(k).view(bs, -1, self.n_head, self.d_k)
        v = self.w_vs(v).view(bs, -1, self.n_head, self.d_k)
        q = q.transpose(1, 2)
        k = k.transpose(1, 2)
        v = v.transpose(1, 2)

        if mask is not None:
            mask = mask.unsqueeze(1)

        output = self.attention(q, k, v, mask=mask)

        output = output.transpose(1, 2).contiguous().view(bs, -1, self.d_model)
        output = self.fc(output)
        output = self.dropout(output)

        return output


class DecoderScaledDotProductAttention(nn.Module):
    def __init__(self, temperature):
        super().__init__()
        self.temperature = temperature
        self.INF = float("inf")

    def forward(self, q, k, v, mask=None):
        attn = torch.matmul(q, k.transpose(2, 3)) / self.temperature
        if mask is not None:
            mask = mask.eq(0)
            attn = attn.masked_fill(mask, -self.INF)
            attn = torch.softmax(attn, dim=-1).masked_fill(mask, 0.0)
        else:
            attn = torch.softmax(attn, dim=-1)
        output = torch.matmul(attn, v)
        return output


class PositionwiseFeedForward(nn.Module):
    def __init__(self, d_model, d_ff, dropout=0.1):
        super().__init__()
        self.w_1 = nn.Linear(d_model, d_ff)
        self.act = nn.GELU()
        self.w_2 = nn.Linear(d_ff, d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        output = self.w_2(self.act(self.w_1(x)))
        output = self.dropout(output)
        return output


class PositionalEncoding(nn.Module):
    def __init__(self, d_model, max_len=5000):
        super().__init__()
        assert d_model % 2 == 0
        pe = torch.zeros(max_len, d_model, requires_grad=False)
        position = torch.arange(0, max_len).unsqueeze(1).float()
        div_term = torch.exp(
            torch.arange(0, d_model, 2).float()
            * -(torch.log(torch.tensor(10000.0)).item() / d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0)
        self.register_buffer("pe", pe)

    def forward(self, x):
        length = x.size(1)
        return self.pe[:, :length].clone().detach()
